

/* Sets up a gateway making thrusterallocation code by DNVGL
 * accessible for python utilizing py4j as gateway between.
 *
 * TODO: Move parameteres to python.
 */
// Import libraries
import py4j.GatewayServer;

import com.marinecyb.vessel.nbm.thrustallocation.core.Allocation;
import com.marinecyb.vessel.nbm.thrustallocation.data.AngleSector;
import com.marinecyb.vessel.nbm.thrustallocation.data.Position;
import com.marinecyb.vessel.nbm.thrustallocation.data.Tau;
import com.marinecyb.vessel.nbm.thrustallocation.data.InputValues;
import com.marinecyb.vessel.nbm.thrustallocation.data.ReturnValues;
import com.marinecyb.vessel.nbm.thrustallocation.data.ThrusterType;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class TA_StarAndPortDisabled {

    private final Allocation allocation;
    private final InputValues inputs;

    public static void main(String[] args) {
        // Start Gateway
        GatewayServer gatewayServer = new GatewayServer(new TA_StarAndPortDisabled());
        gatewayServer.start();
        System.out.println("Gateway Server Started");
    }

    public TA_StarAndPortDisabled() {

        // Setup
        int numThr = 1;
        int numRud = 0;
        allocation = new Allocation(numThr, numRud);

        inputs = new InputValues();
        double[] tMax = {14};
        double[] tMin = {-5.5};
        inputs.setThrustMax(tMax);
        inputs.setThrustMin(tMin);
        inputs.setThrustMaxInit(tMax); // Must be same as ThrustMax
        inputs.setThrustMinInit(tMin); // Must be same as ThrustMin

        inputs.setFixedAngles(new double[]{0.0});
        inputs.setFixedAngleEnables(new boolean[]{false});
        
        Position[] pos = {new Position(1.0, 0.0),};
        inputs.setPositions(pos);
        inputs.setThrusterTypes(new ThrusterType[]{ThrusterType.AZIMUTH});
        inputs.setRudders(new HashMap());
        inputs.setBiasConfigurations(new HashMap());
        inputs.setRateLimitDirections(new double[]{Math.PI/32});  //radians per step
        inputs.setRateLimitThrust(new double[]{4.0});      // newton per step
        inputs.setRateLimitDirectionsEnable(true);
        inputs.setRateLimitThrustEnable(true);
        inputs.setThrusterReady(new boolean[]{true});  // enable thrusters true/false
        inputs.setEnable(true); // allocation on / off
        //forbidden sectors
        double lower = (-90.0 - 20.0) * Math.PI / 180.0; // start of sector, CCW
        double upper = (-90.0 + 20.0) * Math.PI / 180.0; // end of sector, CW
        AngleSector thr2Sector = new AngleSector(lower, upper);
        List thr2List = new ArrayList<>();
        thr2List.add(thr2Sector);
        lower = (90.0 - 20.0) * Math.PI / 180.0; // start of sector, CCW
        upper = (90.0 + 20.0) * Math.PI / 180.0; // end of sector, CW
        AngleSector thr3Sector = new AngleSector(lower, upper);
        List thr3List = new ArrayList<>();
        thr3List.add(thr3Sector);
        Map<Integer, List<AngleSector>> forbiddenSectors = new HashMap<>(); // map with thruster number as key and list of AngleSector objects as values
        forbiddenSectors.put(2, thr3List);
        inputs.setForbiddenSectors(forbiddenSectors);
        inputs.setForbiddenSectorsEnable(false);
        System.out.println("Thrusteralloc Initiated");
    }

    // Function called from python with Tau_d, returns forces and directions
    public double[] runTA(double Xd, double Yd, double Nd) {
		
        // Calculate u
        double[] tMax = {14};
        Tau tau = new Tau(Xd, Yd, Nd);
        inputs.setTauDp(tau);
        ReturnValues returnValues = allocation.iterate(inputs);
        double[] forces_p = returnValues.getForce();   // Thrust commands in percent
        
        //if (forces_p[2] >= 0.0) {
		//	tMax[2]=14;
		//} else {
		//	tMax[2]=5.5;
		//}
		
        double[] forces_n = {0,0,forces_p[0]*tMax[0]/100}; // Thrust commands in newton
        double[] directions = returnValues.getDirection(); // Direction commands in rad
        // Return commanded inputs
        double[] u = {forces_n[0], forces_n[1], forces_n[2], 0, 0, directions[0]}; // Return format {F1, F2, F3, a1, a2, a3}
        return u;
    }
}
