#!/usr/bin/python
import random
import numpy as np
import pandas as pd
import rospy
import os
import time
from sklearn.preprocessing import MinMaxScaler
from keras.models import Sequential
from keras.optimizers import RMSprop
from custom_msgs.msg import CornerTestStatus,inertialSensor,NorthEastHeading,podAngle, gnssRMC, gpsFIX, SternThrusterSetpoints, bowControl, ThrusterFaults
from geometry_msgs.msg import Point
from keras.models import load_model



class actuatorfault_LSTM(object):

	def __init__(self):
		# Create skeleton dataframe for storing mesaurements to be used for fault diagnosis
		self.predictionDF = pd.DataFrame(columns=['Eta_d_N', 'Eta_d_E', 'Eta_d_Heading', 
		'Eta_ref_N','Eta_ref_E', 'Eta_ref_Heading', 'pod_angleInput_port', 'pod_angleInput_star',
		'speed_mps', 'course', 'roll', 'pitch','heave', 'rollRate', 'pitchRate', 'heaveSpeed','star_effort_input',
		'port_effort_input', 'Tau_PID_N', 'Tau_PID_E','Tau_PID_Heading', 'ResidualAttitudeRoll',
		'ResidualAttitudePitch','ResidualAttitudeYaw','ErrorNorth','ErrorNorth_integrated','ErrorNorth_dot',
		'ErrorEast','ErrorEast_integrated','ErrorEast_dot','ErrorPsi','ErrorPsi_integrated','ErrorPsi_dot',
		'position_bow', 'throttle_bow'])

		self.eta_d_DF = pd.DataFrame(columns=['Eta_d_N', 'Eta_d_E', 'Eta_d_Heading'])
		self.eta_ref_DF = pd.DataFrame(columns=['Eta_ref_N','Eta_ref_E', 'Eta_ref_Heading'])
		self.pod_angle_DF = pd.DataFrame(columns=['pod_angleInput_port', 'pod_angleInput_star'])
		self.sog_cog_DF = pd.DataFrame(columns=['speed_mps', 'course'])
		self.VRU_DF = pd.DataFrame(columns=['roll', 'pitch','heave', 'rollRate', 'pitchRate', 'heaveSpeed'])
		self.stern_thruster_DF = pd.DataFrame(columns=['star_effort_input','port_effort_input'])
		self.tau_PID_DF = pd.DataFrame(columns=['Tau_PID_N', 'Tau_PID_E','Tau_PID_Heading'])
		self.residualsAttitude_DF = pd.DataFrame(columns=['ResidualAttitudeRoll','ResidualAttitudePitch','ResidualAttitudeYaw'])
		self.tau_Error_North_DF = pd.DataFrame(columns=['ErrorNorth','ErrorNorth_integrated','ErrorNorth_dot'])
		self.tau_Error_East_DF = pd.DataFrame(columns=['ErrorEast','ErrorEast_integrated','ErrorEast_dot'])
		self.tau_Error_Psi_DF = pd.DataFrame(columns=['ErrorPsi','ErrorPsi_integrated','ErrorPsi_dot'])
		self.bowcontrol_DF = pd.DataFrame(columns=['position_bow', 'throttle_bow'])


		# Init important variables
		self.running = 1
		self.input_dim = 35
		self.timesteps = 1
		self.batch_size_c = 14
		self.batch_constant = 2
		self.epochs = 7
		self.batchfull = [0,0,0,0,0,0,0,0,0,0,0,0]
		self.predict_threshold = 0.5
		self.spinning_freq = 2 # How many seconds shall pass between each actuator fault check
		self.spin_iterate = 0

		#Publish results
		self.fault_results_pub = rospy.Publisher("/actuator_faults_predicted", ThrusterFaults, queue_size=1)

	def ros_loop(self):
		while (self.running == 1):
			rospy.Subscriber("/eta_d", NorthEastHeading, self.eta_d_cb, queue_size=10)
			rospy.Subscriber("/eta_ref", NorthEastHeading, self.eta_ref_CB, queue_size=10)
			rospy.Subscriber("/pod_angle_input", podAngle, self.podangleCB, queue_size=10)
			rospy.Subscriber("/sog_cog", gnssRMC, self.sog_cog_CB, queue_size=10)
			rospy.Subscriber("/virtualInertial1/pre", inertialSensor, self.imuCB, queue_size=10)
			rospy.Subscriber('/stern_thruster_setpoints', SternThrusterSetpoints, self.stern_cb, queue_size=10)
			rospy.Subscriber('/tau_PID', NorthEastHeading, self.tau_PID_cb, queue_size=10)
			rospy.Subscriber('/observer/residualsAttitude', Point, self.attituderesidualsCB, queue_size=10)
			rospy.Subscriber('/tau_Error_North', NorthEastHeading, self.tau_ErrorN_cb, queue_size=10)
			rospy.Subscriber('/tau_Error_East', NorthEastHeading, self.tau_ErrorE_cb, queue_size=10)
			rospy.Subscriber('/tau_Error_Psi', NorthEastHeading, self.tau_ErrorPSI_cb, queue_size=10)
			rospy.Subscriber('/bow_control', bowControl, self.bowCB, queue_size=10)
			self.check_spin()
			rospy.sleep(0.1)
		rospy.spin()


	def model_initialize(self):
		num_datapoints = 0
		iterate = 1
		total_epochs = self.epochs * len(datasets)
		self.epochs_done = 0
		set_names = ['NoFaults','bow_freeze_400','port_freeze_400']
		for k in datasets:
			break_flag = 0

			num_datapoints = datasets[k].shape[0] + num_datapoints


			split_num = 0.7
			train_elements = 0
			test_elements = 0

			while (self.variable_split == 1):
				if(split_num > 0.8):
					break
				split = int(len(datasets[k])*split_num)
				train_elements = split
				test_elements = len(datasets[k])-split
				if ((train_elements%(self.batch_size_c-self.batch_constant) == 0) and (test_elements%(self.batch_size_c-self.batch_constant) == 0)):
					print("found splitting number ", split)
					break
				split_num = split_num + 0.00001
			if(self.variable_split == 0):
				if(break_flag):
					break_flag = 0
					break
				deletedRows = 0
				split = int(len(datasets[k])*split_num)
				train_elements = split
				test_elements = len(datasets[k])-(split)
				while True:
					if ((train_elements%(self.batch_size_c-self.batch_constant) == 0) and (test_elements%(self.batch_size_c-self.batch_constant) == 0)):
						#print("Sets are now divisble by batch size - deleted " + str(deletedRows) +" rows")
						break_flag = 1
						break
					datasets[k].drop((datasets[k].shape[0]-1),inplace = True)
					split = int(len(datasets[k])*split_num)
					train_elements = split
					test_elements = len(datasets[k])-(split)
					deletedRows = deletedRows + 1

			x = datasets[k].iloc[:, :-6] # Prediction variables
			y = datasets[k].iloc[:, -6:] #

			x_train, x_test, y_train, y_test = x[:split], x[split:], y[:split], y[split:]


			# Feature Scaling

			scaler = StandardScaler()
			x_train = scaler.fit_transform(x_train)
			x_test = scaler.transform(x_test)

			# LSTM expect a 3D matrix, so we reshape the training data with an extra
			# singular dimension to satisfy the dimension requirements
			x_train = np.reshape(x_train, (x_train.shape[0],1,x_train.shape[1]))
			x_test = np.reshape(x_test, (x_test.shape[0],1,x_test.shape[1]))

			self.x_testsets[k] = x_test
			self.y_testsets[k] = y_test


			if(hasattr(self,'model_')):
				self.model_.reset_states()
			else:
				self.model_ = Sequential()
				self.model_.add(LSTM(self.batch_size_c, return_sequences=False, batch_size = (self.batch_size_c-self.batch_constant),stateful=True,
							   input_shape=(self.timesteps, self.input_dim)))  # returns a sequence of vectors of dimension 32
				self.model_.add(Dense(512, activation='linear'))
				self.model_.add(LeakyReLU(alpha=0.3))
				self.model_.add(Dense(512, activation='linear'))
				self.model_.add(LeakyReLU(alpha=0.3))
				self.model_.add(Dense(512, activation='linear'))
				self.model_.add(LeakyReLU(alpha=0.3))
				self.model_.add(Dense(512, activation='linear'))
				self.model_.add(LeakyReLU(alpha=0.3))
				self.model_.add(Dense(32, activation='linear'))
				self.model_.add(LeakyReLU(alpha=0.3))
				self.model_.add(Dense(32, activation='linear'))
				self.model_.add(LeakyReLU(alpha=0.3))
				self.model_.add(Dense(4, activation='sigmoid')) #sigmoid paa siste e bra
				opt = RMSprop( lr=0.01 )
				self.model_.compile(loss='binary_crossentropy', optimizer=opt, metrics=['binary_accuracy'])
			history = np.zeros(self.epochs)
			for i in range(self.epochs):
				acc = self.model_.fit(x_train, y_train, epochs=1, batch_size=(self.batch_size_c-self.batch_constant), verbose=0, shuffle=False)
				self.epochs_done = self.epochs_done + 1
				if(self.epochs_done/total_epochs > 0.1*iterate):
					print("Percent done: ",int(100.0*self.epochs_done/total_epochs))
					iterate = iterate + 1
				history[i] = acc.history['loss'][0]
				self.model_.reset_states()




	def model_reload(self, filepath):
		self.model_ = load_model(filepath)
		print("Previous actuator fault model_ detected")

	def shutdown_imminent(self): # Save/dump model_ when the central ROS node is shut down
		self.running = 0
		os.chdir('/home/ubuntudynamo/Desktop/Revolt/src/ml_pack')
		filename = "newLSTM_model_" + str(int(time.time())) + "_.p"
		self.model_.save(filename)
		print("ActuatorNeuralNetwork model has been saved!")

	def check_spin(self):
		self.spin_iterate = self.spin_iterate + 1
		if(int(self.spin_iterate/10) >= self.spinning_freq):
			results = self.check_actuator_faults()
			if not(results is None): 
				print(results[-1][0],results[-1][1],results[-1][2],results[-1][3])
				msg = ThrusterFaults()
				msg.port_norotation = 0
				msg.port_frozen = results[-1][1]
				msg.bow_norotation = 0
				msg.bow_frozen = results[-1][2]
				msg.star_norotation = 0
				msg.star_frozen = results[-1][3]
				self.fault_results_pub.publish(msg)


	def check_actuator_faults(self):
		try:
			self.batchfull.index(0)
			#rospy.logwarn("The measurement containers do not have enough measurements to fill a batch!")
		except ValueError:
			if(sum(self.batchfull) == len(self.batchfull)):
				b = self.batch_size_c-self.batch_constant
				data = {'Eta_d_N': self.Eta_d_N[-b:], 'Eta_d_E': self.Eta_d_E[-b:],'Eta_d_Heading': self.Eta_d_Heading[-b:],
				'Eta_ref_N': self.Eta_ref_N[-b:], 'Eta_ref_E': self.Eta_ref_E[-b:], 'Eta_ref_Heading': self.Eta_ref_Heading[-b:],
				'pod_angleInput_port': self.pod_angleInput_port[-b:], 'pod_angleInput_star': self.pod_angleInput_star[-b:],
				'speed_mps': self.speed_mps[-b:], 'course': self.course[-b:],
				'roll': self.roll[-b:], 'pitch': self.pitch[-b:],'heave': self.heave[-b:], 'rollRate': self.rollRate[-b:], 'pitchRate': self.pitchRate[-b:], 'heaveSpeed': self.heaveSpeed[-b:],
				'star_effort_input': self.star_effort_input[-b:],'port_effort_input': self.port_effort_input[-b:],
				'Tau_PID_N': self.Tau_PID_N[-b:], 'Tau_PID_E': self.Tau_PID_E[-b:],'Tau_PID_Heading': self.Tau_PID_Heading[-b:],
				'ResidualAttitudeRoll':self.ResidualAttitudeRoll[-b:],'ResidualAttitudePitch':self.ResidualAttitudePitch[-b:],'ResidualAttitudeYaw':self.ResidualAttitudeYaw[-b:],
				'ErrorNorth':self.ErrorNorth[-b:],'ErrorNorth_integrated':self.ErrorNorth_integrated[-b:],'ErrorNorth_dot':self.ErrorNorth_dot[-b:],
				'ErrorEast':self.ErrorEast[-b:],'ErrorEast_integrated':self.ErrorEast_integrated[-b:],'ErrorEast_dot':self.ErrorEast_dot[-b:],
				'ErrorPsi':self.ErrorPsi[-b:],'ErrorPsi_integrated':self.ErrorPsi_integrated[-b:],'ErrorPsi_dot':self.ErrorPsi_dot[-b:],
				'position_bow':self.position_bow[-b:],'throttle_bow':self.throttle_bow[-b:]}

				
				measurements = pd.DataFrame(data = data)
				self.predictionDF = self.predictionDF.append(measurements,ignore_index=True)

				scaler = MinMaxScaler(feature_range=(0, 1))
				scaler = scaler.fit(self.predictionDF)
				x_scaled = scaler.transform(self.predictionDF)
				x_scaled = np.reshape(x_scaled, (1,x_scaled.shape[0],x_scaled.shape[1]))

				actuatorfaults = self.model_.predict(x_scaled,batch_size=(self.batch_size_c - self.batch_constant))
				
				for i in range(0,actuatorfaults.shape[0]):
					for j in range(0,actuatorfaults.shape[1]):
						max_val = 0
						max_ind = 0
						for k in range(0,actuatorfaults.shape[2]):
							if (max_val < actuatorfaults[i,j,k]):
								max_ind = k
								max_val = actuatorfaults[i,j,k]
						actuatorfaults[i,j,:] = False
						if(max_val != 0):
							actuatorfaults[i,j,max_ind] = True
				print(actuatorsfaults)
				return actuatorfaults
			else: 
				rospy.logwarn("Thruster ANN: Not enough full measurement containers to run predict")


	# Callback fucntions -- These extract data from messages and put them in measurement containers #
	def eta_d_cb(self,data):
		if(hasattr(self,'Eta_d_N') == False):
			self.Eta_d_N = []
			self.Eta_d_E = []
			self.Eta_d_Heading = []
		if(len(self.Eta_d_N) < self.batch_size_c):
			self.Eta_d_N.append(data.pos_north)
			self.Eta_d_E.append(data.pos_east)
			self.Eta_d_Heading.append(data.pos_heading)
		elif(len(self.Eta_d_N) >= self.batch_size_c):
			del self.Eta_d_N[0]
			del self.Eta_d_E[0]
			del self.Eta_d_Heading[0]
			self.Eta_d_N.append(data.pos_north)
			self.Eta_d_E.append(data.pos_east)
			self.Eta_d_Heading.append(data.pos_heading)
			self.batchfull[0] = 1
	def eta_ref_CB(self,data):
		if(hasattr(self,'Eta_ref_N') == False):
			self.Eta_ref_N = []
			self.Eta_ref_E = []
			self.Eta_ref_Heading = []
		if(len(self.Eta_ref_N) < self.batch_size_c):
			self.Eta_ref_N.append(data.pos_north)
			self.Eta_ref_E.append(data.pos_east)
			self.Eta_ref_Heading.append(data.pos_heading)
		elif(len(self.Eta_ref_N) >= self.batch_size_c):
			del self.Eta_ref_N[0]
			del self.Eta_ref_E[0]
			del self.Eta_ref_Heading[0]
			self.Eta_ref_N.append(data.pos_north)
			self.Eta_ref_E.append(data.pos_east)
			self.Eta_ref_Heading.append(data.pos_heading)
			self.batchfull[1] = 1
	def podangleCB(self,data):
		if(hasattr(self,'pod_angleInput_star') == False):
			self.pod_angleInput_port = []
			self.pod_angleInput_star = []
		if(len(self.pod_angleInput_star) < self.batch_size_c):
			self.pod_angleInput_port.append(data.port)
			self.pod_angleInput_star.append(data.star)
		elif(len(self.pod_angleInput_port) >= self.batch_size_c):
			del self.pod_angleInput_port[0]
			del self.pod_angleInput_star[0]
			self.pod_angleInput_port.append(data.port)
			self.pod_angleInput_star.append(data.star)
			self.batchfull[2] = 1
	def sog_cog_CB(self,data):
		if(hasattr(self,'speed_mps') == False):
			self.speed_mps = []
			self.course = []
		if(len(self.speed_mps) < self.batch_size_c):
			self.speed_mps.append(data.speed_mps)
			self.course.append(data.course)
		elif(len(self.speed_mps) >= self.batch_size_c):
			del self.speed_mps[0]
			del self.course[0]
			self.speed_mps.append(data.speed_mps)
			self.course.append(data.course)
			self.batchfull[3] = 1
	def imuCB(self,data):
		if(hasattr(self,'roll') == False):
			self.roll = []
			self.pitch = []
			self.heave = []
			self.rollRate = []
			self.pitchRate = []
			self.heaveSpeed = []
		if(len(self.roll) < self.batch_size_c):
			self.roll.append(data.roll)
			self.pitch.append(data.pitch)
			self.heave.append(data.heave)
			self.rollRate.append(data.rollRate)
			self.pitchRate.append(data.pitchRate)
			self.heaveSpeed.append(data.heaveSpeed)
		elif(len(self.roll) >= self.batch_size_c):
			del self.roll[0]
			del self.pitch[0]
			del self.heave[0]
			del self.rollRate[0]
			del self.pitchRate[0]
			del self.heaveSpeed[0]
			self.roll.append(data.roll)
			self.pitch.append(data.pitch)
			self.heave.append(data.heave)
			self.rollRate.append(data.rollRate)
			self.pitchRate.append(data.pitchRate)
			self.heaveSpeed.append(data.heaveSpeed)
			self.batchfull[4] = 1
	def stern_cb(self,data):
		if(hasattr(self,'star_effort_input') == False):
			self.star_effort_input = []
			self.port_effort_input = []
		if(len(self.star_effort_input) < self.batch_size_c):
			self.star_effort_input.append(data.star_effort)
			self.port_effort_input.append(data.port_effort)
		elif(len(self.star_effort_input) >= self.batch_size_c):
			del self.star_effort_input[0]
			del self.port_effort_input[0]
			self.star_effort_input.append(data.star_effort)
			self.port_effort_input.append(data.port_effort)
			self.batchfull[5] = 1
	def tau_PID_cb(self,data):
		if(hasattr(self,'Tau_PID_N') == False):
			self.Tau_PID_N = []
			self.Tau_PID_E = []
			self.Tau_PID_Heading = []
		if(len(self.Tau_PID_N) < self.batch_size_c):
			self.Tau_PID_N.append(data.pos_north)
			self.Tau_PID_E.append(data.pos_east)
			self.Tau_PID_Heading.append(data.pos_heading)
		elif(len(self.Tau_PID_N) >= self.batch_size_c):
			del self.Tau_PID_N[0]
			del self.Tau_PID_E[0]
			del self.Tau_PID_Heading[0]
			self.Tau_PID_N.append(data.pos_north)
			self.Tau_PID_E.append(data.pos_east)
			self.Tau_PID_Heading.append(data.pos_heading)
			self.batchfull[6] = 1
	def attituderesidualsCB(self,data):
		if(hasattr(self,'ResidualAttitudeRoll') == False):
			self.ResidualAttitudeRoll = []
			self.ResidualAttitudePitch = []
			self.ResidualAttitudeYaw = []
		if(len(self.ResidualAttitudeRoll) < self.batch_size_c):
			self.ResidualAttitudeRoll.append(data.x)
			self.ResidualAttitudePitch.append(data.y)
			self.ResidualAttitudeYaw.append(data.z)
		elif(len(self.ResidualAttitudeRoll) >= self.batch_size_c):
			del self.ResidualAttitudeRoll[0]
			del self.ResidualAttitudePitch[0]
			del self.ResidualAttitudeYaw[0]
			self.ResidualAttitudeRoll.append(data.x)
			self.ResidualAttitudePitch.append(data.y)
			self.ResidualAttitudeYaw.append(data.z)
			self.batchfull[7] = 1
	def tau_ErrorN_cb(self,data):
		if(hasattr(self,'ErrorNorth') == False):
			self.ErrorNorth = []
			self.ErrorNorth_integrated = []
			self.ErrorNorth_dot = []
		if(len(self.ErrorNorth) < self.batch_size_c):
			self.ErrorNorth.append(data.pos_north)
			self.ErrorNorth_integrated.append(data.pos_east)
			self.ErrorNorth_dot.append(data.pos_heading)
		elif(len(self.ErrorNorth) >= self.batch_size_c):
			del self.ErrorNorth[0]
			del self.ErrorNorth_integrated[0]
			del self.ErrorNorth_dot[0]
			self.ErrorNorth.append(data.pos_north)
			self.ErrorNorth_integrated.append(data.pos_east)
			self.ErrorNorth_dot.append(data.pos_heading)
			self.batchfull[8] = 1
	def tau_ErrorE_cb(self,data):
		if(hasattr(self,'ErrorEast') == False):
			self.ErrorEast = []
			self.ErrorEast_integrated = []
			self.ErrorEast_dot = []
		if(len(self.ErrorEast) < self.batch_size_c):
			self.ErrorEast.append(data.pos_north)
			self.ErrorEast_integrated.append(data.pos_east)
			self.ErrorEast_dot.append(data.pos_heading)
		elif(len(self.ErrorEast) >= self.batch_size_c):
			del self.ErrorEast[0]
			del self.ErrorEast_integrated[0]
			del self.ErrorEast_dot[0]
			self.ErrorEast.append(data.pos_north)
			self.ErrorEast_integrated.append(data.pos_east)
			self.ErrorEast_dot.append(data.pos_heading)
			self.batchfull[9] = 1
	def tau_ErrorPSI_cb(self,data):
		if(hasattr(self,'ErrorPsi') == False):
			self.ErrorPsi = []
			self.ErrorPsi_integrated = []
			self.ErrorPsi_dot = []
		if(len(self.ErrorPsi) < self.batch_size_c):
			self.ErrorPsi.append(data.pos_north)
			self.ErrorPsi_integrated.append(data.pos_east)
			self.ErrorPsi_dot.append(data.pos_heading)
		elif(len(self.ErrorPsi) >= self.batch_size_c):
			del self.ErrorPsi[0]
			del self.ErrorPsi_integrated[0]
			del self.ErrorPsi_dot[0]
			self.ErrorPsi.append(data.pos_north)
			self.ErrorPsi_integrated.append(data.pos_east)
			self.ErrorPsi_dot.append(data.pos_heading)
			self.batchfull[10] = 1
			
	# Position and throttle are shifted on purpose!!
	def bowCB(self,data):
		if(hasattr(self,'position_bow') == False):
			self.position_bow = []
			self.throttle_bow = []
		if(len(self.position_bow) < self.batch_size_c):
			self.position_bow.append(data.throttle_bow)
			self.throttle_bow.append(data.position_bow)
		elif(len(self.position_bow) >= self.batch_size_c):
			del self.position_bow[0]
			del self.throttle_bow[0]
			self.position_bow.append(data.throttle_bow)
			self.throttle_bow.append(data.position_bow)
			self.batchfull[11] = 1


def main():
	# initiate node
	rospy.init_node('ThrusterFaultNode')
	load_model = 1 # Flag to choose whether to reload an old model_ or train a new one

	ActuatorNeuralNetwork = actuatorfault_LSTM()

	if(load_model == 1): # Reload old machine learning model_ (no training)
		filepath = "/home/ubuntudynamo/Desktop/Revolt/src/ml_pack/insertmodelhere"
		ActuatorNeuralNetwork.model_reload(filepath)
	else: # Train new machine learning model_ on basis of old data (no previous model_)

		# Read CSV file containing initial training data
		os.chdir('/home/ubuntudynamo/Desktop/Revolt/src/ml_pack')

		# Create a dataframe object containing the initial training data
		df_dictionary = {}
		df_dictionary[0] = pd.read_csv('trainingdata_nofaults.csv', encoding = "ISO-8859-1")
		df_dictionary[1] = pd.read_csv('trainingdata_bow_400_freeze.csv', encoding = "ISO-8859-1")
		df_dictionary[2] = pd.read_csv('trainingdata_port_400_freeze.csv', encoding = "ISO-8859-1")
		df_dictionary[3] = pd.read_csv('trainingdata_star_400_freeze.csv', encoding = "ISO-8859-1")
		df_dictionary[4] = pd.read_csv('trainingdata_bow_stuck.csv', encoding = "ISO-8859-1")
		df_dictionary[5] = pd.read_csv('trainingdata_port_stuck.csv', encoding = "ISO-8859-1")
		df_dictionary[6] = pd.read_csv('trainingdata_star_stuck.csv', encoding = "ISO-8859-1")
		ActuatorNeuralNetwork.model_initialize(df_dictionary)
	rospy.on_shutdown(ActuatorNeuralNetwork.shutdown_imminent)
	while(ActuatorNeuralNetwork.running):
		ActuatorNeuralNetwork.ros_loop()



if __name__ == '__main__':
	try:
		main()
	except rospy.ROSInterruptException:
		pass
