#!/usr/bin/python
import rospy
import time
import numpy as np
from numpy import matrix
from numpy import pi, nan
from numpy.matlib import eye, ones, zeros, diag
from numpy.linalg import norm
from numpy import shape
from ekfLib import *
from scipy.linalg import block_diag as blkdiag
import math

from std_msgs.msg import Float64
from std_msgs.msg import Bool
from custom_msgs.msg import NorthEastHeading, ImuSensor, gnssGGA, gnssHDT, gnssRMC
from custom_msgs.msg import twistEstimate, ekfBias, EKF_Kmat, EKF_Pmat, residualPose, filterStates
from custom_msgs.srv import ChangeToDP
from geometry_msgs.msg import Point, Vector3, Quaternion, Twist
from dynamic_reconfigure.parameter_generator_catkin import *
from dynamic_reconfigure.server import Server
from dp_controller.cfg import resetnedConfig


class EKF:

    def __init__(self):
        print 'EKF initiated'
        self.parameters = self.parameters()
        self.resetFlag = 0

    class parameters:

        def __init__(self):
            self.zG = np.mat([0.0, 0.0, 9.81])
            # Known bias
            self.gravityE = np.mat([0, 0, 0]).T 
            # Observation Matrix
        
            gnss = np.bmat([np.eye(3), np.zeros((3, 13))])
            attitude = np.bmat([np.zeros((4,3)),np.eye(4),np.zeros((4,9))])
            velo_imu = np.bmat([np.zeros((3,7)),np.eye(3),np.zeros((3,6))])
            velo_gnss = np.bmat([np.zeros((2,7)),np.eye(2),np.zeros((2,7))])
            self.H = np.bmat([[gnss], [attitude],[velo_imu],[velo_gnss]])
            
            
            # Standard Deviations
            std_accel = 0.2 # 0.20
            std_gyros = 0.05
            std_pose_GNSS = 1.2
            std_velo_GNSS = 0.15
            
            # Measurement Covariances
            self.R_PoseGNSS  = np.eye(3)*std_pose_GNSS**2
            self.R_VeloGNSS = np.eye(2)*std_velo_GNSS**2
            self.R_Gyros = np.eye(4)*std_gyros**2
            self.R_Accel = 0*np.eye(3)*std_accel**2
            
            # Measurement Covariance Matrix
            self.R = blkdiag(self.R_PoseGNSS, self.R_Gyros,self.R_Accel,self.R_VeloGNSS)

            # Process Covariances 
            qPosition = (0.455*np.eye(3))**2 #0.455
            qAttitude = (0.05*np.eye(4))**2 #0.035
            qVelocity = (0.22*np.eye(3))**2
            qAccelBias = np.eye(3)**2
            qGyroBias = np.eye(3)**2
            # Process Covariance Matrix 
            self.Q  = blkdiag(qPosition,qAttitude,qVelocity,qAccelBias,qGyroBias) 

    def NoPrevStateNorState(self, currentTime, filterAttitude):
        self.deltaT = 0
        if hasattr(self, 'zG') and hasattr(self, 'zC'):
            filterAttitude[2] = self.zC
            q = ypr2quat(filterAttitude)
            self.prevstates = matrix([
                float(self.zG[0]),
                float(self.zG[1]),
                float(self.zG[2]),
                float(np.asscalar(q[0])),
                float(np.asscalar(q[1])),
                float(np.asscalar(q[2])),
                float(np.asscalar(q[3])),
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                ]).T
                

            self.states = matrix([
                float(self.zG[0]),
                float(self.zG[1]),
                float(self.zG[2]),
                float(np.asscalar(q[0])),
                float(np.asscalar(q[1])),
                float(np.asscalar(q[2])),
                float(np.asscalar(q[3])),
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                ]).T
            self.velocityBody = np.zeros((3,1))
            #self.P_k = np.zeros((16,16))
            self.P_k = np.eye(16)
            self.P_k[1,1] = 0.627
            self.P_k[2,2] = 0.627
            self.P_k[3,3] = 0.63
            self.P_k[4,4] = 0.001
            self.P_k[5,5] = 0.001
            self.P_k[6,6] = 0.001
            self.P_k[7,7] = 0.001
            self.P_k[8,8] = 0.0337
            self.P_k[9,9] = 0.0337
            self.P_k[10,10] = 0.1345




            self.Innovation = zeros((16,1))
            self.kalmanGain = np.zeros((16,10))
            print("State and prev state reset")
            self.prevTime = currentTime

    def ResetEKFObserver(self, currentTime, imuData, filterAttitude): 
        EKFObserverReset = 0
        i = 0
        while EKFObserverReset < 0.5 and i < 20:
            if hasattr(self, 'zG') and hasattr(self, 'zC'):
                self.NoPrevStateNorState(currentTime, filterAttitude)
                self.resetFlag = 1
                EKFObserverReset = 1
                print('EKFObserver successfully reset.')
            else:
                i = i+1
                time.sleep(0.1)

        if EKFObserverReset < 0.5:
            print('EKFObserver not reset. Please activate DP mode once more')

    def compassMea(self, zC): 
        if not np.isnan(zC):
            self.zC = zC
        else:
            if hasattr(self, 'zC'):
                del self.zC
    def sogcogMea(self, sog,cog):
        if not (np.isnan(sog) and np.isnan(cog)):
            self.zS = sog
            self.zCG = cog*(np.pi/180)
        else:
            if hasattr(self, 'zS'):
                del self.zS
                del self.zCG
    def gpsMea(self, zG):  # Fyller inn GNSS maling
        if not np.isnan(zG[0]) or not np.isnan(zG[1]) or not np.isnan(zG[2]):
            self.zG = zG
        else:
            if hasattr(self, 'zG'):
                del self.zG
                self.zG = zG


    def iterate(self, currentTime, imuData, filterAttitude, gnssData={}):
        
        if not hasattr(self, 'prevstates') or not hasattr(self, 'prevTime'):
            self.NoPrevStateNorState(currentTime, filterAttitude)
            print('init state (observer)')
            return
        param = self.parameters        
        # Calculate time passed since last iteration

        deltaT = currentTime - self.prevTime
        self.prevTime = currentTime
        self.deltaT = deltaT
        

        statesEst_pre = self.states
        

        if self.resetFlag == 1:
            print("Reset at start checkpoint")
            return
        # Compute Rotation matrix based on last quaternion measurement
        q_states = statesEst_pre[3:7]
        q_states_normalized = quatNormalizer(q_states)
        RotationQuat = rotationmatrix_quat(q_states_normalized)


        # Extract Measurements         
        LinearAcceleration_m = imuData[0:3]
        AngularVelocity_m = imuData[3:6]
        Velocity_m_imu = RotationQuat.dot(imuData[6:9])
        quaternion_m_imu = ypr2quat(filterAttitude)
        #Normalize
        quaternion_m_imu = quatNormalizer(quaternion_m_imu)
        
        if hasattr(self, 'zG'):
            posMeasuredN = self.zG[0]
            posMeasuredE = self.zG[1]
            posMeasuredD = self.zG[2]
            VelocityN_m_GNSS = np.matrix([self.zS*np.cos(self.zCG)])
            VelocityE_m_GNSS = np.matrix([self.zS*np.sin(self.zCG)])
            measurements = np.bmat([[posMeasuredN],[posMeasuredE],[posMeasuredD],[quaternion_m_imu],[Velocity_m_imu],[VelocityN_m_GNSS],[VelocityE_m_GNSS]])
        else:
            posMeasuredN = statesEst_pre[0]
            posMeasuredE = statesEst_pre[1]
            posMeasuredD = statesEst_pre[2]
            rospy.logwarn("No GNSS measurement")


        # Prediction Step 
        
        position_dot = statesEst_pre[7:10]
        position_bar = statesEst_pre[0:3] + position_dot*deltaT

        
        quaternion_dot = quaternionDot(statesEst_pre[3:7],(AngularVelocity_m-statesEst_pre[13:16]))
        self.qdot = quaternion_dot
        quaternion_bar = statesEst_pre[3:7] + quaternion_dot*deltaT #quatMultiply(statesEst_pre[3:7],quaternion_dot*deltaT) #quatMultiply(quaternion_dot*deltaT,statesEst_pre[3:7])
        #Normalize quaternion
        quaternion_bar = quatNormalizer(quaternion_bar)

        velo_dot = (RotationQuat.dot(LinearAcceleration_m - statesEst_pre[10:13]) + param.gravityE)
        velo_bar = statesEst_pre[7:10] + velo_dot*deltaT


        if self.resetFlag == 1:
            self.NoPrevStateNorState(currentTime, filterAttitude)
            print("Reset at second checkpoint")
            return

        ## State Propagation Matrix f(X,u)
        fX = np.bmat([[position_bar],[quaternion_bar],[velo_bar],[np.zeros((3,1))],[np.zeros((3,1))]])
        
        # Jacobian of State Propagation Matrix => d(fX)/dX
        FX_jacobi = statePropagJacobian(deltaT,imuData[3:6],statesEst_pre[3:7],imuData[0:3],statesEst_pre[10:13],statesEst_pre[13:16])
        
        P_k = self.P_k
        P_bar_kPlus = np.linalg.multi_dot([FX_jacobi,P_k,FX_jacobi.T]) + param.Q

       
        if self.resetFlag == 1:
            self.NoPrevStateNorState(currentTime, filterAttitude)
            print("Reset at third checkpoint")
            return        
        
        
        # Predicted States
        
        stateBar = fX

        ## UPDATE 
        ## Residual Covariance => H*P*H^T + R
        
        S = np.linalg.multi_dot([param.H,P_bar_kPlus,param.H.T]) + param.R
        
        ## Kalman Gain Matrix => P*H^T / H*P*H^T + R
        
        K = np.linalg.multi_dot([P_bar_kPlus,param.H.T,S.I])
        #K[3,:] = 0.0
        self.kalmanGain = K
        ## State Estimate Update
        
        # Quaternion State Estimate Update (Own procedure due to preservation of the unit quaternion)
        
        quatInnovationTerm = quatInnovation(measurements[3:7],stateBar[3:7])
        quatHat = quatStateUpdate(stateBar[3:7],K[4:7,4:7],quatInnovationTerm)
        
        # Regular State Estimation Update (all states except quaternion)
        self.Innovation = measurements-param.H.dot(stateBar)
        
        
        stateHat  = stateBar + K*(self.Innovation)
        # Overwrite non-normalized quaternion with the quatHat
        stateHat[3:7] = quatHat
        
        
        ## Updated State Covariance 
        
        P_kPlus = P_bar_kPlus - np.linalg.multi_dot([K,S,K.T])
        
        if self.resetFlag == 1:
            del self.kalmanGain
            del self.Innovation
            self.NoPrevStateNorState(currentTime, filterAttitude)
            print("Reset at last checkpoint")
            return
            
        ## Advance to the next step
        
        self.P_k = P_kPlus
        self.states = stateHat
        
        

 

class Observer:

    def __init__(self):
        print 'Observer initiated'
        self.lat0 = get_param('~lat0', 63.439175163)
        self.lon0 = get_param('~lon0', 10.400329186)
        self.alt0 = get_param('~alt0', -0.682)

        self.ekf = EKF()
        self.pubRawGNSS = 0 # Whether to publish unfiltered GNSS data
        # Init publishers
        self.pub_eta_GNSS = rospy.Publisher('eta_GNSS', NorthEastHeading, queue_size=1) # This one is used in the control system!
        self.pub_GNSS_unfiltered = rospy.Publisher('GNSS_unfiltered',NorthEastHeading, queue_size=1)
        self.pub_GNSS_actual = rospy.Publisher('GNSS_true', NorthEastHeading, queue_size=1)

        self.pub_sog_cog = rospy.Publisher("sog_cog", gnssRMC, queue_size=1)
        self.pub_NED = rospy.Publisher("ned_origin", gnssGGA, queue_size=1, latch=True)
        self.observedFixPub = rospy.Publisher('observer/fix', Point, queue_size=10)
        self.etaPub = rospy.Publisher("observer/eta", Twist, queue_size=1)
        self.nuPub = rospy.Publisher("observer/nu", Twist, queue_size=1)
        self.observedTwistPub = rospy.Publisher('observer/twist', twistEstimate, queue_size=10)
        self.observedTwistNEDpub = rospy.Publisher('observer/twistNED', twistEstimate, queue_size = 10)
        self.observedEulerPub = rospy.Publisher('observer/orientation', Vector3, queue_size=10)
        self.ekfAccbiaspub = rospy.Publisher('observer/Accbias', Vector3, queue_size=10)
        self.ekfbGyrbiaspub = rospy.Publisher('observer/Gyrbias', Vector3, queue_size=10)
        self.ekfdeltaTPub = rospy.Publisher('observer/deltaT', Vector3, queue_size=10)
        self.vruPub = rospy.Publisher('observer/vru', Vector3, queue_size=10)
        self.gpsfix = rospy.Publisher('observer/gpsfix', Point, queue_size=10)
        self.gpsyaw = rospy.Publisher('observer/gpsyaw', Vector3, queue_size=10)
        self.statePub = rospy.Publisher('observer/states',filterStates,queue_size=1)
        self.pMat = rospy.Publisher('observer/EKF_pMat', EKF_Pmat, queue_size=10)
        self.KMat = rospy.Publisher('observer/EKF_KMat', EKF_Kmat, queue_size=10)


        self.dropped = 0

        # Init callbacks
        FIXss = rospy.Subscriber("/signal/GNSS_post", gnssGGA, self.FIXcb)
        FIXssactual = rospy.Subscriber("vectorVS330/fix", gnssGGA, self.FIXcb_actual)
        IMUss = rospy.Subscriber("xsens/imu", ImuSensor, self.IMUcb)
        HDTss = rospy.Subscriber("vectorVS330/heading", gnssHDT, self.HDTcb)
        VELss = rospy.Subscriber("vectorVS330/velocity", gnssRMC, self.SOGCOGcb)
        DPss = rospy.Service("change_to_DP", ChangeToDP, self.parameter_callback_changeToDP)        
        rospy.loginfo("Initialized")

        # Publish reference lla to guidance law node
        ned_origin = gnssGGA()
        ned_origin.latitude = self.lat0
        ned_origin.longitude = self.lon0
        self.pub_NED.publish(ned_origin)





    def FIXcb(self, data):

        self.latitude = data.latitude
        self.longitude = data.longitude
        self.altitude = data.altitude
        if np.isnan(self.latitude) or np.isnan(self.longitude) or np.isnan(self.altitude):
            rospy.logwarn('No GPS measurements for observer')
            return
        if not hasattr(self, 'heading') or np.isnan(self.heading):
            rospy.logwarn('No compass measurements for observer')
            return
        if not hasattr(self, 'roll') or not hasattr(self, 'pitch') or np.isnan(self.roll) or np.isnan(self.pitch):
            rospy.logwarn('No attitude measurements for observer')
            return

        self.NED = lla2ned(self.latitude, self.longitude, self.altitude, self.lat0, self.lon0, self.alt0)

        # Do lever arm comp
        self.NED = levercomp(self.NED, self.heading, self.roll, self.pitch)
        #print("NED:")
        #print(self.NED)
        self.ekf.gpsMea(self.NED)


        # Publish
        
        eta_unfiltered = NorthEastHeading()
        eta_unfiltered.pos_heading = self.heading * (180 / pi)
        eta_unfiltered.pos_north = self.NED[0]
        eta_unfiltered.pos_east = self.NED[1]
        self.pub_GNSS_unfiltered.publish(eta_unfiltered)
        
        if(self.pubRawGNSS == 1):
            self.pub_eta_GNSS.publish(eta_unfiltered)
        
    def FIXcb_actual(self, data):
        if np.isnan(data.latitude) or np.isnan(data.longitude) or np.isnan(data.altitude):
            rospy.logwarn('No GPS measurements for observer')
            return
        if not hasattr(self, 'heading') or np.isnan(self.heading):
            rospy.logwarn('No compass measurements for observer')
            return
        if not hasattr(self, 'roll') or not hasattr(self, 'pitch') or np.isnan(self.roll) or np.isnan(self.pitch):
            rospy.logwarn('No attitude measurements for observer')
            return
        self.NED = lla2ned(data.latitude, data.longitude, data.altitude, self.lat0, self.lon0, self.alt0)

        # Do lever arm comp
        self.NED = levercomp(self.NED, self.heading, self.roll, self.pitch)

        
        eta = NorthEastHeading()
        eta.pos_heading = self.heading * (180 / pi)
        eta.pos_north = self.NED[0]
        eta.pos_east = self.NED[1]
        self.pub_GNSS_actual.publish(eta)
    
    
    ''' Observer subscribes to Speed Over Ground, Course Over ground (no filtering atm) '''
    def SOGCOGcb(self, data):
        self.sog = data.speed_mps
        #rospy.logwarn("SOG: %f COG: %f",data.speed_mps, data.course)
        if(data.course > 180.0):
            self.cog = data.course - 360
        else:
            self.cog = data.course
        
        if np.isnan(self.sog) or np.isnan(self.cog):
            rospy.logwarn('no sog/cog measurment')
            return
        
        self.ekf.sogcogMea(self.sog,self.cog)
        message = gnssRMC()
        message.speed_mps = self.sog
        message.course = self.cog
        self.pub_sog_cog.publish(message)



    def HDTcb(self, data):  # Yaw angle from GPHDT is in degrees and wrapped
        self.heading = data.heading * (np.pi / 180)
        #self.heading = wrapHeading_rad(self.heading)
        self.ekf.compassMea(self.heading)

    def IMUcb(self, data):
        if not hasattr(self, 'heading') or np.isnan(self.heading):
            rospy.logwarn('No compass measurements for observer')
            return

        self.roll = data.orientation.roll * (np.pi / 180)
        self.pitch = data.orientation.pitch * (np.pi / 180)
        yaw = data.orientation.yaw * (np.pi / 180)
        self.imuData = np.array([[data.linear_acceleration.x],
                                 [data.linear_acceleration.y],
                                 [data.linear_acceleration.z],
                                 [data.angular_velocity.x * (np.pi / 180)],
                                 [data.angular_velocity.y * (np.pi / 180)],
                                 [data.angular_velocity.z * (np.pi / 180)],
                                 [data.linear_velocity.x],
                                 [data.linear_velocity.y],
                                 [data.linear_velocity.z]])
        self.filterAttitude = matrix([self.roll,
                                      self.pitch, yaw]).T
        self.timestamp = rospy.Time.now().secs+rospy.Time.now().nsecs * 10 ** (-9)

        if self.ekf.resetFlag == 1:
            self.ekf.NoPrevStateNorState(self.timestamp, self.filterAttitude)
            self.ekf.resetFlag = 0
            return
        self.ekf.iterate(self.timestamp, self.imuData,self.filterAttitude)
        if self.ekf.resetFlag == 1:
            self.ekf.NoPrevStateNorState(self.timestamp, self.filterAttitude)
            self.ekf.resetFlag = 0
            return
        if hasattr(self.ekf, 'states'):
            # print quat2ypr(self.ekf.state[3:7])*180/pi
            #print('ekf state: ', self.ekf.state)
            fix = Point()
            fix.x = self.ekf.states[0]
            fix.y = self.ekf.states[1]
            fix.z = self.ekf.states[2]
            
            
            orientation = Vector3()
            eulAngle = quat2ypr(self.ekf.states[3:7]) * (180 / pi)
            orientation.x = eulAngle[0]
            orientation.y = eulAngle[1]
            orientation.z = eulAngle[2]
            # Twist msgs
            twist = twistEstimate()
            velNED = self.ekf.states[7:10]
            q = self.ekf.states[3:7]
            # Normalize
            q = quatNormalizer(q)
            
            R = rotationMatrixEuler(eulAngle[0] * pi / 180, eulAngle[1] * pi / 180, eulAngle[2] * pi / 180)
            T = angularVelTransformation_quat(q)
            velBody = R.T.dot(velNED)
            
            #if hasattr(self.ekf, 'qdota'):
                #angularVelBody = BODYangularRatesfromQuat(q,self.ekf.qdot)*(180/np.pi)
            #else:
            angularVelBody = T.T.dot(quatDifferentialVik(q, (self.imuData[3:6] - self.ekf.states[13:16])))*231 # The scaling factor has been found by studying actual yaw rate values and those produced by QuatDifferential
            
            if self.ekf.resetFlag == 1:
                self.ekf.resetFlag = 0
                print("Reset at imuCB checkpoint")
                return
            twist.linear_velocity.x = velBody[0]
            twist.linear_velocity.y = velBody[1]
            twist.linear_velocity.z = velBody[2]
            twist.angular_velocity.x = angularVelBody[0]
            twist.angular_velocity.y = angularVelBody[1]
            twist.angular_velocity.z = angularVelBody[2]
            
            # Twist (NED)
            twistNED = twistEstimate()
            
            twistNED.linear_velocity.x = self.ekf.states[7]
            twistNED.linear_velocity.y = self.ekf.states[8]
            twistNED.linear_velocity.z = self.ekf.states[9]
            
            angularVelNED = quatDifferentialVik(q, (self.imuData[3:6] - self.ekf.states[13:16])) # QuatDifferential function needs revision, way too small values!
            twistNED.angular_velocity.x = angularVelNED[0]
            twistNED.angular_velocity.y = angularVelNED[1]
            twistNED.angular_velocity.z = angularVelNED[2]
            
            # Eta (N, E, D, roll, pitch, yaw)
            eta = Twist()
            eta.linear.x = self.ekf.states[0]
            eta.linear.y = self.ekf.states[1]
            eta.linear.z = self.ekf.states[2]
            eta.angular.x = eulAngle[0]
            eta.angular.y = eulAngle[1]
            eta.angular.z = eulAngle[2]


            if(self.pubRawGNSS == 0):
                gnss = NorthEastHeading()
                #NED = self.ekf.states[0:3]
                #heading = eulAngle[2]*(np.pi/180)
                #gnssFiltered = levercomp(NED, heading, self.roll, self.pitch)
                #gnss.pos_north = gnssFiltered[0]
                #gnss.pos_east = gnssFiltered[1]
                gnss.pos_north = self.ekf.states[0]
                gnss.pos_east = self.ekf.states[1]
                gnss.pos_heading = eulAngle[2]
                self.pub_eta_GNSS.publish(gnss)
            # Nu (u, v, w, p,q,r)
            nu = Twist()
            #nu.linear.x = velBody[0]
            #nu.linear.y = velBody[1]
            #nu.linear.z = velBody[2]

            roll_rad = eulAngle[0]*(pi/180)
            pitch_rad = eulAngle[1]*(pi/180)
            yaw_rad = eulAngle[2]*(pi/180)
            
            nu.linear.x = velBody[0]*(np.cos(yaw_rad)*np.cos(pitch_rad)) + velBody[1]*(np.cos(yaw_rad)*np.sin(pitch_rad)*np.sin(roll_rad) - np.sin(yaw_rad)*np.cos(roll_rad)) + velBody[2]*(np.sin(yaw_rad)*np.sin(roll_rad)+np.cos(yaw_rad)*np.cos(roll_rad)*np.sin(pitch_rad))
            nu.linear.y = velBody[0]*(np.sin(yaw_rad)*np.cos(pitch_rad)) + velBody[1]*(np.cos(yaw_rad)*np.cos(roll_rad) + np.sin(yaw_rad)*np.sin(roll_rad)*np.sin(pitch_rad)) + velBody[2]*(np.sin(yaw_rad)*np.sin(pitch_rad)*np.cos(roll_rad)-np.cos(yaw_rad)*np.sin(roll_rad))
            nu.linear.z = velBody[2]
            nu.angular.x = angularVelBody[0]
            nu.angular.y = angularVelBody[1]
            nu.angular.z = angularVelBody[2]

            # bias
            Accbias = Vector3()
            Accbias.x = self.ekf.states[10]
            Accbias.y = self.ekf.states[11]
            Accbias.z = self.ekf.states[12]
            Gyrbias = Vector3()
            Gyrbias.x = self.ekf.states[13]
            Gyrbias.y = self.ekf.states[14]
            Gyrbias.z = self.ekf.states[15]
            
            states = filterStates()
            states.N = self.ekf.states[0]
            states.E = self.ekf.states[1]
            states.D = self.ekf.states[2]
            states.w = self.ekf.states[3]
            states.q1 = self.ekf.states[4]
            states.q2 = self.ekf.states[5]
            states.q3 = self.ekf.states[6]
            states.N_dot = self.ekf.states[7]
            states.E_dot = self.ekf.states[8]
            states.D_dot = self.ekf.states[9]
            states.acc_bias_x = self.ekf.states[10]
            states.acc_bias_y = self.ekf.states[11]
            states.acc_bias_z = self.ekf.states[12]
            states.gyro_bias_z = self.ekf.states[13]
            states.gyro_bias_z = self.ekf.states[14]
            states.gyro_bias_z = self.ekf.states[15]
            



            
            # Fill P Mat msg
            pMat = EKF_Pmat()
            pMat.P_11 = self.ekf.P_k[0, 0]
            pMat.P_22 = self.ekf.P_k[1, 1]
            pMat.P_33 = self.ekf.P_k[2, 2]
            pMat.P_44 = self.ekf.P_k[3, 3]
            pMat.P_55 = self.ekf.P_k[4, 4]
            pMat.P_66 = self.ekf.P_k[5, 5]          
            pMat.P_77 = self.ekf.P_k[6, 6]
            pMat.P_88 = self.ekf.P_k[7, 7]
            pMat.P_99 = self.ekf.P_k[8, 8]
            pMat.P_1010 = self.ekf.P_k[9, 9]
            pMat.P_1111 = self.ekf.P_k[10, 10]
            pMat.P_1212 = self.ekf.P_k[11, 11]
            pMat.P_1313 = self.ekf.P_k[12, 12]
            pMat.P_1414 = self.ekf.P_k[13, 13]
            pMat.P_1515 = self.ekf.P_k[14, 14]
            pMat.P_1616 = self.ekf.P_k[15, 15]

            # Fill K Mat msg
            KMat = EKF_Kmat()
            KMat.K_11 = self.ekf.kalmanGain[0, 0]
            KMat.K_22 = self.ekf.kalmanGain[1,1]
            KMat.K_33 = self.ekf.kalmanGain[2,2]
            KMat.K_44 = self.ekf.kalmanGain[3,3]
            KMat.K_55 = self.ekf.kalmanGain[4,4]
            KMat.K_66 = self.ekf.kalmanGain[5,5]
            KMat.K_77 = self.ekf.kalmanGain[6,6]
            KMat.K_88 = self.ekf.kalmanGain[7,7]
            KMat.K_99 = self.ekf.kalmanGain[8,8]
            KMat.K_1010 = self.ekf.kalmanGain[9,9]



            self.pMat.publish(pMat)
            self.KMat.publish(KMat)
            self.ekfAccbiaspub.publish(Accbias)
            self.ekfbGyrbiaspub.publish(Gyrbias)
            self.observedFixPub.publish(fix)
            self.nuPub.publish(nu)
            self.etaPub.publish(eta)
            self.observedEulerPub.publish(orientation)
            self.observedTwistPub.publish(twist)
            self.observedTwistNEDpub.publish(twistNED)
            self.statePub.publish(states)



    def parameter_callback(self, mode, level):
        if mode.set == 1:
            rospy.set_param('~lon0', self.longitude)
            rospy.set_param('~lat0', self.latitude)
            rospy.set_param('~alt0', self.altitude)
            print self.longitude
            print self.latitude
            print self.altitude
            self.lat0 = get_param('~lat0', 63.439206)
            self.lon0 = get_param('~lon0', 10.399675)
            self.alt0 = get_param('~alt0', 39)

            self.NED = lla2ned(self.latitude, self.longitude, self.altitude, self.lat0, self.lon0, self.alt0)
            # Publishes NED origin to the Guidance system
            ned_origin = gnssGGA()
            ned_origin.latitude = self.lat0
            ned_origin.longitude = self.lon0
            self.pub_NED.publish(ned_origin)
            mode.set = 0
        return mode
    
    def parameter_callback_changeToDP(self,level):
        rospy.set_param('~lon0', self.longitude)
        rospy.set_param('~lat0', self.latitude)
        rospy.set_param('~alt0', self.altitude)
        print self.longitude
        print self.latitude
        print self.altitude
        self.lat0 = get_param('~lat0', 63.439206)
        self.lon0 = get_param('~lon0', 10.399675)
        self.alt0 = get_param('~alt0', 39)
        self.NED = lla2ned(self.latitude, self.longitude, self.altitude, self.lat0, self.lon0, self.alt0)
        # Publishes NED origin to the Guidance system
        ned_origin = gnssGGA()
        ned_origin.latitude = self.lat0
        ned_origin.longitude = self.lon0
        self.pub_NED.publish(ned_origin)

        print 'Reseting EKF observer'
        self.ekf.ResetEKFObserver(self.timestamp, self.imuData ,self.filterAttitude)
        

        return True

def get_param(name, default):
    try:
        v = rospy.get_param(name)
        rospy.loginfo("Found parameter: %s, value: %s" % (name, str(v)))
    except KeyError:
        v = default
        rospy.logwarn("Cannot find value for parameter: %s, assigning "
                      "default: %s" % (name, str(v)))
    return v


# Initiate
def main():

    # initiate node
    rospy.init_node('ObserverNode')
    observer = Observer()
    srv = Server(resetnedConfig, observer.parameter_callback)

    rospy.spin()

    # while not rospy.is_shutdown():
    #   main()


if __name__ == '__main__':
    try:
        main()
    except rospy.ROSInterruptException:
        pass
