# Install scipy: pip install git+https://github.com/scipy/scipy.git #?Linux?
# Install scipy: https://sourceforge.net/projects/scipy/files/scipy/0.14.0/ #Windows 64bit python 2.7
# Install numpy: pip install numpy

import numpy as np
from scipy.linalg import expm  # Used in discretise_without_u
from math import pow          # Used in lla2ned


def angularDifference(controlZ, minuend, subtrahend):
    c = np.mod(controlZ, 2 * np.pi)
    minuendShifted = np.mod(minuend - c, 2 * np.pi)
    subtrahendShifted = np.mod(subtrahend - c, 2 * np.pi)
    zOut = minuendShifted - subtrahendShifted
    z2 = subtrahendShifted - minuendShifted

    if zOut > minuendShifted:
        zOut = z2

    return zOut

def quaternionDot(q,omega):
    # q_dot = 0.5 * deltaT * Q_omega * q where Q is a matrix using ang. rates as argument
    Q_w = np.bmat([[np.zeros((1,1)),omega.T],[-omega,skewsym(omega)]])
    q_dot = 0.5*(-Q_w).dot(q) #0.5*deltaT*Q_w.dot(q) 27/4
    
    return q_dot

def quaternionDot_Fossen(q,omega):
    gamma = 100
    
    Tq = angularVelTransformation_quat(q)
    qTq = np.asscalar(q.T.dot(q))
    convergence_term = ((gamma/2)*(1-qTq))*q
    q_dot = Tq.dot(omega) + convergence_term
    return q_dot
    
def statePropagJacobian_FX(deltaT,R):
    
    FX_jacobi = np.zeros((16,16))
    
    # f1
    FX_jacobi[0,0] = 1
    FX_jacobi[1,1] = 1
    FX_jacobi[2,2] = 1
    FX_jacobi[0,7] = deltaT
    FX_jacobi[1,8] = deltaT
    FX_jacobi[2,9] = deltaT
    
    #f2 
    #quat2RotationMatrix_Sola
    FX_jacobi[4,7] = 1
    FX_jacobi[5,8] = 1
    FX_jacobi[6,9] = 1
    FX_jacobi[4,10] = -deltaT*R[0,0]
    FX_jacobi[4,11] = -deltaT*R[0,1]
    FX_jacobi[4,12] = -deltaT*R[0,2]
    FX_jacobi[5,10] = -deltaT*R[1,0]
    FX_jacobi[5,11] = -deltaT*R[1,1]
    FX_jacobi[5,12] = -deltaT*R[1,2]
    FX_jacobi[6,10] = -deltaT*R[2,0]
    FX_jacobi[6,11] = -deltaT*R[2,1]
    FX_jacobi[6,12] = -deltaT*R[2,2]
    
    #f3
    FX_jacobi[8,3] = 1
    FX_jacobi[9,4] = 1
    FX_jacobi[10,5] = 1
    FX_jacobi[11,6] = 1
    
    
    return FX_jacobi

def statePropagJacobian_FX_updated(deltaT,omega,q,acc,acc_bias,g_bias):
    
    
    W = q[0]
    q_1 = q[1]
    q_2 = q[2]
    q_3 = q[3]
    
    omega_1 = omega[0]
    omega_2 = omega[1]
    omega_3 = omega[2]
    
    acc_X = acc[0]
    acc_Y = acc[1]
    acc_Z = acc[2]
    acc_bias_X = acc_bias[0]
    acc_bias_Y = acc_bias[1]
    acc_bias_Z = acc_bias[2]
    
    FX_jacobi = np.zeros((16,16))
    
    # f1
    FX_jacobi[0,0] = 0
    FX_jacobi[1,1] = 0
    FX_jacobi[2,2] = 0
    FX_jacobi[0,7] = 1
    FX_jacobi[1,8] = 1
    FX_jacobi[2,9] = 1
    
    #f2 
    #quat2RotationMatrix_Sola bias?
    FX_jacobi[3,4] = -omega_1
    FX_jacobi[3,5] = -omega_2
    FX_jacobi[3,6] = -omega_3
    
    FX_jacobi[4,3] = omega_1
    FX_jacobi[4,5] = omega_3
    FX_jacobi[4,6] = -omega_2
    
    FX_jacobi[5,3] = omega_2
    FX_jacobi[5,4] = -omega_3
    FX_jacobi[5,6] = omega_1
    
    FX_jacobi[6,3] = omega_3
    FX_jacobi[6,4] = omega_2
    FX_jacobi[6,5] = -omega_1

    #f3
    FX_jacobi[7,3] = -2*q_2*(acc_Z - acc_bias_Z) + 2*q_3*acc_bias_Y
    FX_jacobi[7,4] = 2*q_2*(acc_Y-acc_bias_Y) + 2*q_3*(acc_Z-acc_bias_Z)
    FX_jacobi[7,5] = 2*q_1*(acc_Y-acc_bias_Y)+4*q_2*acc_bias_X-4*q_2*acc_X - 2*W*(acc_Z - acc_bias_Z)
    FX_jacobi[7,6] = 2*q_1*(acc_Z-acc_bias_Z) + 2*W*acc_bias_Y - 4*q_3*acc_X + 4*q_3*acc_bias_X
    FX_jacobi[7,10] = -1+2*(q_1*q_1 + q_2*q_2)
    FX_jacobi[7,11] = 2*q_3*W-2*q_1*q_2
    FX_jacobi[7,12] = -2*q_1*q_3 + 2*q_2*W
    
    FX_jacobi[8,3] = -2*q_1*(acc_Z - acc_bias_Z) - 2*q_3*(acc_X-acc_bias_X)
    FX_jacobi[8,4] = 2*q_2*(acc_X-acc_bias_X) - 4*q_1*acc_Y + 4*q_1*acc_bias_Y - 2*W*(acc_Z - acc_bias_Z)
    FX_jacobi[8,5] = 2*q_1*(acc_X - acc_bias_X) + 2*q_3*(acc_Z - acc_bias_Z)
    FX_jacobi[8,10] = -2*q_1*q_2 + 2*q_3*W
    FX_jacobi[8,11] = 2*q_1*q_1 + 2*q_3*q_3 - 1
    FX_jacobi[8,12] = 2*q_1*W - 2*q_2*q_3
    
    FX_jacobi[9,3] = 2*q_1*(acc_Y-acc_bias_Y)-2*q_2*(acc_X - acc_bias_X)
    FX_jacobi[9,4] = 2*W*(acc_Y-acc_bias_Y) + 2*q_3*(acc_X-acc_bias_X) + 4*q_1*(acc_bias_Z-acc_Z)
    FX_jacobi[9,5] = -2*W*(acc_X-acc_bias_X) + 2*q_3*(acc_Y-acc_bias_Y) + 4*q_2*(acc_bias_Z - acc_Z)
    FX_jacobi[9,6] = 2*q_1*(acc_X-acc_bias_X) + 2*q_2*(acc_Y-acc_bias_Y)
    FX_jacobi[9,10] = -2*q_1*q_2 + 2*q_2*W
    FX_jacobi[9,11] = -2*q_1*W - 2*q_2*q_3
    FX_jacobi[9,12] = 2*q_1*q_1 + 2*q_2*q_2 - 1     
     
    
    
    return FX_jacobi

def statePropagJacobian(deltaT,omega,q,acc,acc_bias,g_bias):
    
    
    W = q[0]
    q_1 = q[1]
    q_2 = q[2]
    q_3 = q[3]
    
    omega_1 = omega[0]
    omega_2 = omega[1]
    omega_3 = omega[2]

    omega_1_bias = g_bias[0]
    omega_2_bias = g_bias[1]
    omega_3_bias = g_bias[2]
    
    acc_X = acc[0]
    acc_Y = acc[1]
    acc_Z = acc[2]
    acc_bias_X = acc_bias[0]
    acc_bias_Y = acc_bias[1]
    acc_bias_Z = acc_bias[2]
    
    FX_jacobi = np.zeros((16,16))
    
    # f1
    FX_jacobi[0,0] = 1
    FX_jacobi[1,1] = 1
    FX_jacobi[2,2] = 1
    FX_jacobi[0,7] = deltaT
    FX_jacobi[1,8] = deltaT
    FX_jacobi[2,9] = deltaT
    
    #f2 
    #quat2RotationMatrix_Sola bias?
    FX_jacobi[3,3] = 1    
    FX_jacobi[3,4] = 0.5*deltaT*(omega_1-omega_1_bias)
    FX_jacobi[3,5] = 0.5*deltaT*(omega_2-omega_2_bias)
    FX_jacobi[3,6] = 0.5*deltaT*(omega_3-omega_3_bias)
    FX_jacobi[3,13] = 0.5*deltaT*q_1
    FX_jacobi[3,14] = 0.5*deltaT*q_2
    FX_jacobi[3,15] = 0.5*deltaT*q_3
    
    
    
    FX_jacobi[4,3] = 0.5*deltaT*(omega_1-omega_1_bias)
    FX_jacobi[4,4] = 1
    FX_jacobi[4,5] = 0.5*deltaT*(omega_3-omega_3_bias)
    FX_jacobi[4,6] = -0.5*deltaT*(omega_2-omega_2_bias)
    FX_jacobi[4,13] = -0.5*deltaT*W
    FX_jacobi[4,14] = 0.5*deltaT*q_3
    FX_jacobi[4,15] = -0.5*deltaT*q_2
    
    
    FX_jacobi[5,3] = 0.5*deltaT*(omega_2-omega_2_bias)
    FX_jacobi[5,4] = -0.5*deltaT*(omega_3-omega_3_bias)
    FX_jacobi[5,5] = 1
    FX_jacobi[5,6] = 0.5*deltaT*(omega_1-omega_1_bias)
    FX_jacobi[5,13] = -0.5*deltaT*q_3
    FX_jacobi[5,14] = -0.5*deltaT*W
    FX_jacobi[5,15] = 0.5*deltaT*q_1    
    
    
    FX_jacobi[6,3] = 0.5*deltaT*(omega_3-omega_3_bias)
    FX_jacobi[6,4] = 0.5*deltaT*(omega_2-omega_2_bias)
    FX_jacobi[6,5] = -0.5*deltaT*(omega_1-omega_1_bias)
    FX_jacobi[6,6] = 1
    FX_jacobi[6,13] = 0.5*deltaT*q_2
    FX_jacobi[6,14] = -0.5*deltaT*q_1
    FX_jacobi[6,15] = -0.5*deltaT*W   

    #f3
    FX_jacobi[7,3] = -2*deltaT*q_2*(acc_Z - acc_bias_Z) + 2*deltaT*q_3*acc_bias_Y
    FX_jacobi[7,4] = 2*deltaT*q_2*(acc_Y-acc_bias_Y) + 2*deltaT*q_3*(acc_Z-acc_bias_Z)
    FX_jacobi[7,5] = 2*deltaT*q_1*(acc_Y-acc_bias_Y)+4*q_2*deltaT*acc_bias_X-4*q_2*deltaT*acc_X - 2*W*deltaT*(acc_Z - acc_bias_Z)
    FX_jacobi[7,6] = 2*q_1*deltaT*(acc_Z-acc_bias_Z) + 2*W*deltaT*acc_bias_Y - 4*deltaT*q_3*acc_X + 4*deltaT*q_3*acc_bias_X
    FX_jacobi[7,7] = 1
    FX_jacobi[7,10] = -deltaT+2*deltaT*(q_1*q_1 + q_2*q_2)
    FX_jacobi[7,11] = 2*deltaT*q_3*W-2*q_1*q_2*deltaT
    FX_jacobi[7,12] = -2*deltaT*q_1*q_3 + 2*deltaT*q_2*W
    
    FX_jacobi[8,3] = -2*deltaT*q_1*(acc_Z - acc_bias_Z) - 2*deltaT*q_3*(acc_X-acc_bias_X)
    FX_jacobi[8,4] = 2*deltaT*q_2*(acc_X-acc_bias_X) - 4*deltaT*q_1*acc_Y + 4*deltaT*q_1*acc_bias_Y - 2*W*deltaT*(acc_Z - acc_bias_Z)
    FX_jacobi[8,5] = 2*deltaT*q_1*(acc_X - acc_bias_X) + 2*deltaT*q_3*(acc_Z - acc_bias_Z)
    FX_jacobi[8,8] = 1
    FX_jacobi[8,10] = -2*deltaT*q_1*q_2 + 2*deltaT*q_3*W
    FX_jacobi[8,11] = 2*deltaT*q_1*q_1 + 2*deltaT*q_3*q_3 - deltaT
    FX_jacobi[8,12] = 2*deltaT*q_1*W - 2*deltaT*q_2*q_3
    
    FX_jacobi[9,3] = 2*deltaT*q_1*(acc_Y-acc_bias_Y)-2*deltaT*q_2*(acc_X - acc_bias_X)
    FX_jacobi[9,4] = 2*deltaT*W*(acc_Y-acc_bias_Y) + 2*deltaT*q_3*(acc_X-acc_bias_X) + 4*deltaT*q_1*(acc_bias_Z-acc_Z)
    FX_jacobi[9,5] = -2*deltaT*W*(acc_X-acc_bias_X) + 2*deltaT*q_3*(acc_Y-acc_bias_Y) + 4*deltaT*q_2*(acc_bias_Z - acc_Z)
    FX_jacobi[9,6] = 2*deltaT*q_1*(acc_X-acc_bias_X) + 2*deltaT*q_2*(acc_Y-acc_bias_Y)
    FX_jacobi[9,9] = 1
    FX_jacobi[9,10] = -2*deltaT*q_1*q_2 + 2*deltaT*q_2*W
    FX_jacobi[9,11] = -2*deltaT*q_1*W - 2*deltaT*q_2*q_3
    FX_jacobi[9,12] = 2*deltaT*q_1*q_1 + 2*deltaT*q_2*q_2 - deltaT    
     
    
    
    return FX_jacobi

def rotationMatrixEuler(roll, pitch, yaw):

    r = np.zeros((3, 3))

    r[0, 0] = np.cos(yaw) * np.cos(pitch)
    r[0, 1] = -np.sin(yaw) * np.cos(roll) + np.cos(yaw) * np.sin(pitch) * np.sin(roll)
    r[0, 2] = np.sin(yaw) * np.sin(roll) + np.cos(yaw) * np.cos(roll) * np.sin(pitch)
    r[1, 0] = np.sin(yaw) * np.cos(pitch)
    r[1, 1] = np.cos(yaw) * np.cos(roll) + np.sin(roll) * np.sin(pitch) * np.sin(yaw)
    r[1, 2] = -np.cos(yaw) * np.sin(roll) + np.sin(pitch) * np.sin(yaw) * np.cos(roll)
    r[2, 0] = -np.sin(pitch)
    r[2, 1] = np.cos(pitch) * np.sin(roll)
    r[2, 2] = np.cos(pitch) * np.cos(roll)

    return r

def ned2lla(lat_deg, long_deg, altitude, lat0_deg, long0_deg, altitude0):
    # From vik 2012 - page 4 - table 1.1
    re = 6378137  # [m]
    rp = 6356752  # [m]
    deg2rad = np.pi / 180
    rad2deg = 180 / np.pi



    lat_rad = lat_deg * deg2rad
    long_rad = long_deg * deg2rad
    lat0_rad = lat0_deg * deg2rad
    long0_rad = long0_deg * deg2rad

    # N in vik
    rad_curve = (re ** 2) / (np.sqrt((re ** 2) * (np.cos(lat_rad) ** 2) + (rp ** 2) * (np.sin(lat_rad) ** 2)))

    # LLA to ECEF
    xe = (rad_curve + altitude) * np.cos(lat_rad) * np.cos(long_rad)
    ye = (rad_curve + altitude) * np.cos(lat_rad) * np.sin(long_rad)
    ze = ((pow(rp, 2) / pow(re, 2)) * rad_curve + altitude) * np.sin(lat_rad)

    xe0 = (rad_curve + altitude0) * np.cos(lat0_rad) * np.cos(long0_rad)
    ye0 = (rad_curve + altitude0) * np.cos(lat0_rad) * np.sin(long0_rad)
    ze0 = ((pow(rp, 2) / pow(re, 2)) * rad_curve + altitude0) * np.sin(lat0_rad)

    # ECEF to NED
    cosPhi = np.cos(lat0_rad)
    sinPhi = np.sin(lat0_rad)
    cosLambda = np.cos(long0_rad)
    sinLambda = np.sin(long0_rad)

    dist = cosLambda * xe + sinLambda * ye
    uNorth = -sinPhi * dist + cosPhi * ze
    vEast = -sinLambda * xe + cosLambda * ye
    wDown = -(cosPhi * dist + sinPhi * ze)

    # ECEF0 to NED0
    cosPhi0 = cosPhi
    sinPhi0 = sinPhi
    cosLambda0 = cosLambda
    sinLambda0 = sinLambda

    dist0 = cosLambda0 * xe0 + sinLambda0 * ye0
    vEast0 = -sinLambda0 * xe0 + cosLambda0 * ye0
    wDown0 = -(cosPhi0 * dist0 + sinPhi0 * ze0)
    uNorth0 = -sinPhi0 * dist0 + cosPhi0 * ze0

    # Compute difference
    N = uNorth - uNorth0  # [m]
    E = vEast - vEast0   # [m]
    D = wDown - wDown0   # [m]

    NED = np.bmat(np.array([[N], [E], [D]]))

    return NED


def lla2ned(lat_deg, long_deg, altitude, lat0_deg, long0_deg, altitude0):
    deg2rad = np.pi / 180
    rad2deg = 180 / np.pi

    lat_rad = lat_deg * deg2rad
    long_rad = long_deg * deg2rad
    lat0_rad = lat0_deg * deg2rad
    long0_rad = long0_deg * deg2rad
    # From vik 2012 - page 4 - table 1.1
    re = 6378137  # [m]
    rp = 6356752  # [m]
    # N in vik
    rad_curve = (re ** 2) / (np.sqrt((re ** 2) * (np.cos(lat_rad) ** 2) + (rp ** 2) * (np.sin(lat_rad) ** 2)))

    # LLA to ECEF
    xe = (rad_curve + altitude) * np.cos(lat_rad) * np.cos(long_rad)
    ye = (rad_curve + altitude) * np.cos(lat_rad) * np.sin(long_rad)
    ze = ((pow(rp, 2) / pow(re, 2)) * rad_curve + altitude) * np.sin(lat_rad)

    xe0 = (rad_curve + altitude0) * np.cos(lat0_rad) * np.cos(long0_rad)
    ye0 = (rad_curve + altitude0) * np.cos(lat0_rad) * np.sin(long0_rad)
    ze0 = ((pow(rp, 2) / pow(re, 2)) * rad_curve + altitude0) * np.sin(lat0_rad)

    # ECEF to NED
    cosPhi = np.cos(lat0_rad)
    sinPhi = np.sin(lat0_rad)
    cosLambda = np.cos(long0_rad)
    sinLambda = np.sin(long0_rad)

    dist = cosLambda * xe + sinLambda * ye
    uNorth = -sinPhi * dist + cosPhi * ze
    vEast = -sinLambda * xe + cosLambda * ye
    wDown = -(cosPhi * dist + sinPhi * ze)

    # ECEF0 to NED0
    cosPhi0 = cosPhi
    sinPhi0 = sinPhi
    cosLambda0 = cosLambda
    sinLambda0 = sinLambda

    dist0 = cosLambda0 * xe0 + sinLambda0 * ye0
    vEast0 = -sinLambda0 * xe0 + cosLambda0 * ye0
    wDown0 = -(cosPhi0 * dist0 + sinPhi0 * ze0)
    uNorth0 = -sinPhi0 * dist0 + cosPhi0 * ze0

    # Compute difference
    N = uNorth - uNorth0  # [m]
    E = vEast - vEast0   # [m]
    D = wDown - wDown0   # [m]

    NED = np.bmat(np.array([[N], [E], [D]]))

    return NED

def quatNormalizer(q):
    norm = np.sqrt(q[0]**2+q[1]**2+q[2]**2+q[3]**2)
    if(abs(1-norm) < 1e-08):
        return q
    else:
        q_n = np.zeros((4,1))
        q_n[0], q_n[1], q_n[2], q_n[3] = q[0]/norm, q[1]/norm, q[2]/norm, q[3]/norm
        
        return q_n

def quat2ypr(q):
    # if ROS complains about this function, it might be time to update numpy :)


    qw = q[0]
    qx = q[1]
    qy = q[2]
    qz = q[3]

    roll = np.arctan2(2*(qy*qz + qw*qx),qw*qw-qx*qx-qy*qy+qz*qz)
    pitch = np.arcsin(-2*(qx*qz - qw*qy))
    yaw = np.arctan2(2*(qx*qy + qw*qz), qw*qw + qx*qx-qy*qy -qz*qz)
    
    output = np.bmat([[roll], [pitch], [yaw]]).A

    return output
    
def quatRotationMatrix_body2NED(q):
    # See Fossen 2011 page 29
    
    row1 = np.array([(1-2*(q[2]**2+q[3]**2)),(2*(q[1]*q[2]-q[3]*q[0])),(2*(q[1]*q[3]+q[2]*q[0]))]).T 
    row2 = np.array([(2*(q[1]*q[2]+q[3]*q[0])),(1-2*(q[1]**2+q[3]**2)),(2*(q[2]*q[3]-q[1]*q[0]))]).T
    row3 = np.array([(2*(q[1]*q[3]-q[2]*q[0])),(2*(q[2]*q[3]+q[1]*q[0])),(1-2*(q[1]**2+q[2]**2))]).T
    
    return np.bmat([[row1],[row2],[row3]])
    
def quat2RotationMatrix_Sola(q):
    # See Sola 2017 page 25
    
    row1 = np.array([(q[0]**2 + q[1]**2 - q[2]**2 - q[3]**2),(2*(q[1]*q[2]-q[0]*q[3])),(2*(q[1]*q[3]+q[0]*q[2]))]).T 
    row2 = np.array([(2*(q[1]*q[2]+q[3]*q[0])),(q[0]**2 - q[1]**2 + q[2]**2 - q[3]**2),(2*(q[2]*q[3]-q[0]*q[1]))]).T
    row3 = np.array([(2*(q[1]*q[3]-q[2]*q[0])),(2*(q[2]*q[3]+q[1]*q[0])),(q[0]**2 - q[1]**2 - q[2]**2 + q[3]**2)]).T
    
    return np.bmat([[row1],[row2],[row3]])
    
def quat2transformationMatrixAngular_body2NED(q):
    eulers = quat2ypr(q)
    roll = np.asscalar(eulers[0])
    pitch = np.asscalar(eulers[1])
    yaw = np.asscalar(eulers[2])
    
    T = np.array([[1,np.sin(roll)*np.tan(pitch),np.cos(roll)*np.tan(pitch)],
                    [0,np.cos(roll),-np.sin(roll)],
                    [0,(np.sin(roll)/np.cos(pitch)),(np.cos(roll)/np.cos(pitch))]])
    return T

def quat2transformationMatrixAngular_NED2body(q):
    eulers = quat2ypr(q)
    roll = np.asscalar(eulers[0])
    pitch = np.asscalar(eulers[1])
    yaw = np.asscalar(eulers[2])
    
    T = np.array([[1,0,-np.sin(pitch)],
                    [0,np.cos(roll),np.cos(pitch)*np.sin(roll)],
                    [0,-np.sin(roll),np.cos(roll)*np.cos(pitch)]])
    return T

def quatDifferentialSola(deltaTheta):
    deltaThetaNorm = np.linalg.norm(deltaTheta)
    if deltaThetaNorm > 0:
        deltaQ = np.zeros((4, 1))
        deltaQ[0] = np.cos(deltaThetaNorm / 2)
        deltaQ[1:4] = (deltaTheta / deltaThetaNorm) * np.sin(deltaThetaNorm / 2)
    else:
        deltaQ = np.zeros((4, 1))
        deltaQ[0] = 1

    return deltaQ

def BODYangularRatesfromQuat(q,q_dot):
    
    G_matrix = np.array([[-q[1],q[0],q[3],-q[2]],
                         [-q[2],-q[3],q[0],q[1]],
                         [-q[3],q[2],-q[1],q[0]]])
    G_matrix = np.reshape(G_matrix,(3,4))
    
    return 2*G_matrix.dot(q_dot)

def quatDifferentialVik(q, omega):
    eta = q[0]
    epsilon = q[1:4]
    xiMat = np.zeros((4, 3))
    xiMat[0, :] = np.multiply(-0.5, epsilon).T

    sMat = np.zeros((3, 3))
    sMat[0, 1] = -epsilon[2]
    sMat[0, 2] = epsilon[1]
    sMat[1, 0] = epsilon[2]
    sMat[1, 2] = -epsilon[0]
    sMat[2, 0] = -epsilon[1]
    sMat[2, 1] = epsilon[0]

    xiMat[1:4, :] = (np.multiply(eta, np.eye(3)) + sMat) / 2

    qDiff = xiMat.dot(omega)

    return qDiff


def NEDangularRatesfromQuat(q,q_dot):
    
    
    E_matrix = np.array([[-q[1],q[0],-q[3],q[2]],
                         [-q[2],q[3],q[0],-q[1]],
                         [-q[3],-q[2],q[1],q[0]]])
    E_matrix = np.reshape(E_matrix,(3,4))
    
    omega = 2.0*(E_matrix.dot(q_dot))
    return omega

def quatMultiply(p, q):
    p = np.bmat([p])
    q = np.bmat([q])
    p0 = p[0]
    q0 = q[0]

    pB = p[1:4]
    qB = q[1:4]

    r1 = p0 * q0
    r2 = sum(np.multiply(pB, qB))
    r3 = np.multiply(np.bmat([[q0], [q0], [q0]]), pB)
    r4 = np.multiply(np.bmat([[p0], [p0], [p0]]), qB)


    r5 = np.cross(pB.A1, qB.A1)
    r5 = r5.reshape(3, 1)

    ra = r1 - r2
    rb = r3 + r4 + r5

    output = np.bmat([[ra], [rb]]).A
    return output


def quatInnovation(q_measured,q_estimate):
    # Conjugate q_estimate 
    q_estimate_conjugated = np.matrix([[q_estimate[0,0]],[-1*q_estimate[1,0]],[-1*q_estimate[2,0]],[-1*q_estimate[3,0]]])
    quatInnovation = quatMultiply(q_measured, q_estimate_conjugated)
    return quatInnovation[1:4]
    
    
def quatStateUpdate(q_estimate,K_q,quatInnovationTerm):
    KqInnovateTerm = K_q.dot(quatInnovationTerm)
    KqInnovateTerm = np.asscalar(np.sqrt(KqInnovateTerm[0]**2+KqInnovateTerm[1]**2+KqInnovateTerm[2]**2))

    normalizingQuat = np.array([[np.sqrt(1-KqInnovateTerm**2)],[np.asscalar(quatInnovationTerm[0])],[np.asscalar(quatInnovationTerm[1])],[np.asscalar(quatInnovationTerm[2])]])
    normalizingQuat = normalizingQuat.reshape(4,1)

    quatStateUpdated = quatMultiply(q_estimate, normalizingQuat)
    quatStateUpdated = quatNormalizer(quatStateUpdated)
    return quatStateUpdated
def angularVelTransformation_quat(q):

    eta = q[0]
    epsilon = q[1:4]

    T = np.zeros((4, 3))
    T[0, 0] = -epsilon[0]
    T[0, 1] = -epsilon[1]
    T[0, 2] = -epsilon[2]
    T[1, 0] = eta
    T[1, 1] = -epsilon[2]
    T[1, 2] = epsilon[1]
    T[2, 0] = epsilon[2]
    T[2, 1] = eta
    T[2, 2] = -epsilon[0]
    T[3, 0] = -epsilon[1]
    T[3, 1] = epsilon[0]
    T[3, 2] = eta

    T = T.dot(0.5)
    return T

def quatRotate(xIn, q):

    qA = np.zeros((4, 1))
    qA[0] = q[0]
    qA[1:4] = -q[1:4]

    xE = np.zeros((4, 1))
    xE[0] = 0
    xE[1:4] = xIn[0:3]

    qF1 = xE[0] * qA[0] - xE[1] * qA[1] - xE[2] * qA[2] - xE[3] * qA[3]
    qF2 = xE[1] * qA[0] + xE[0] * qA[1] + xE[2] * qA[3] - xE[3] * qA[2]
    qF3 = xE[2] * qA[0] + xE[0] * qA[2] + xE[3] * qA[1] - xE[1] * qA[3]
    qF4 = xE[3] * qA[0] + xE[0] * qA[3] + xE[1] * qA[2] - xE[2] * qA[1]
    qF = np.bmat('qF1; qF2; qF3; qF4').T

    qS1 = q[0] * qF[0] - q[1] * qF[1] - q[2] * qF[2] - q[3] * qF[3]
    qS2 = q[1] * qF[0] + q[0] * qF[1] + q[2] * qF[3] - q[3] * qF[2]
    qS3 = q[2] * qF[0] + q[0] * qF[2] + q[3] * qF[1] - q[1] * qF[3]
    qS4 = q[3] * qF[0] + q[0] * qF[3] + q[1] * qF[2] - q[2] * qF[1]
    qS = np.bmat('qS1 qS2 qS3 qS4').T

    output = qS[1:4]

    return output


def rotationmatrix_quat(q):
    # Fossen page 28 -> R_body2ned = I + 2*w*S + 2*S*S
    eta = q[0]
    epsilon = q[1:4]

    s = np.zeros((3, 3))
    s[0, 1] = -epsilon[2]
    s[0, 2] = epsilon[1]
    s[1, 0] = epsilon[2]
    s[1, 2] = -epsilon[0]
    s[2, 0] = -epsilon[1]
    s[2, 1] = epsilon[0]

    R = np.eye(3) + np.multiply(2 * eta, s) + np.multiply(2, s.dot(s))

    return R


def skewsym(x):
    s = np.zeros((3, 3))
    s[0, 1] = -x[2]
    s[0, 2] = x[1]
    s[1, 0] = x[2]
    s[1, 2] = -x[0]
    s[2, 0] = -x[1]
    s[2, 1] = x[0]

    return s


def ypr2quat(eulerArr):

    roll = eulerArr[0]
    pitch = eulerArr[1]
    yaw = eulerArr[2]

    cR = np.cos(roll / 2)
    sR = np.sin(roll / 2)
    cP = np.cos(pitch / 2)
    sP = np.sin(pitch / 2)
    cY = np.cos(yaw / 2)
    sY = np.sin(yaw / 2)

    q1 = cR * cP * cY + sR * sP * sY
    q2 = sR * cP * cY - cR * sP * sY
    q3 = cR * sP * cY + sR * cP * sY
    q4 = cR * cP * sY - sR * sP * cY
    q = np.bmat([[q1], [q2], [q3], [q4]])

    return q


def wrapHeading_rad(heading):
    """Wraps angle between -pi - pi"""
    x = np.cos(heading)
    y = np.sin(heading)
    return(np.arctan2(y, x))


def levercomp(NED, heading, roll, pitch):
    R = rotationMatrixEuler(roll, pitch, heading)
    DeltaP = np.array([0, 0, -0.2])
    DeltaP = DeltaP.reshape(3, 1)

    NED = NED + R.T.dot(DeltaP)
    return NED

def get_param(name, default):
    try:
        v = rospy.get_param(name)
        rospy.loginfo("Found parameter: %s, value: %s" % (name, str(v)))
    except KeyError:
        v = default
        rospy.logwarn("Cannot find value for parameter: %s, assigning "
                      "default: %s" % (name, str(v)))
    return v


### All functions below are from the NavPy package, as licensed under the BSD license. 

def ned2ecef(ned,lat_ref,lon_ref,alt_ref,latlon_unit='deg',alt_unit='m',model='wgs84'):
    """
    Transform a vector resolved in NED (origin given by lat_ref, lon_ref, and alt_ref)
    coordinates to its ECEF representation. 
    Parameters
    ----------
    ned : {(N,3)} input array, units of meters
    lat_ref : Reference latitude, unit specified by latlon_unit, default in deg
    lon_ref : Reference longitude, unit specified by latlon_unit, default in deg
    alt_ref : Reference altitude, unit specified by alt_unit, default in m
    
    Returns
    -------
    ecef : {(N,3)} array like ned vector, in the ECEF frame, units of meters
    Notes
    -----
    The NED vector is treated as a relative vector, and hence the ECEF representation
    returned is NOT converted into an absolute coordinate.  This means that the 
    magnitude of `ned` and `ecef` will be the same (bar numerical differences).
    
    Examples
    --------
    >>> import navpy
    >>> ned = [0, 0, 1]
    >>> lat_ref, lon_ref, alt_ref = 45.0, -93.0, 250.0 # deg, meters
    >>> ecef = navpy.ned2ecef(ned, lat_ref, lon_ref, alt_ref)
    >>> print("NED:", ned)
    >>> print("ECEF:", ecef)
    >>> print("Notice that 'down' is not same as 'ecef-z' coordinate.")
    """
    lat_ref,N1 = input_check_Nx1(lat_ref)
    lon_ref,N2 = input_check_Nx1(lon_ref)
    alt_ref,N3 = input_check_Nx1(alt_ref)
    
    if( (N1!=1) or (N2!=1) or (N3!=1) ):
        raise ValueError('Reference Location can only be 1')
    
    ned,N = input_check_Nx3(ned)

    ned = ned.T
    
    C = np.zeros((3,3))

    if(latlon_unit=='deg'):
        lat_ref = np.deg2rad(lat_ref)
        lon_ref = np.deg2rad(lon_ref)
    elif(latlon_unit=='rad'):
        pass
    else:
        raise ValueError('Input unit unknown')

    C[0,0]=-np.sin(lat_ref)*np.cos(lon_ref)
    C[0,1]=-np.sin(lat_ref)*np.sin(lon_ref)
    C[0,2]= np.cos(lat_ref)
    
    C[1,0]=-np.sin(lon_ref)
    C[1,1]= np.cos(lon_ref)
    C[1,2]= 0

    C[2,0]=-np.cos(lat_ref)*np.cos(lon_ref)
    C[2,1]=-np.cos(lat_ref)*np.sin(lon_ref)
    C[2,2]=-np.sin(lat_ref)

    # C defines transoformation: ned = C * ecef.  Hence used transpose.
    ecef = np.dot(C.T,ned)
    ecef = ecef.T

    if(N == 1):
        ecef = ecef.reshape(3)

    return ecef
    
def lla2ecef(lat, lon, alt, latlon_unit='deg', alt_unit='m', model='wgs84'):
    """
    Convert Latitude, Longitude, Altitude, to ECEF position
    
    Parameters
    ----------
    lat : {(N,)} array like latitude, unit specified by latlon_unit, default in deg
    lon : {(N,)} array like longitude, unit specified by latlon_unit, default in deg
    alt : {(N,)} array like altitude, unit specified by alt_unit, default in m
    
    Returns
    -------
    ecef : {(N,3)} array like ecef position, unit is the same as alt_unit
    """
    lat,N1 = input_check_Nx1(lat)
    lon,N2 = input_check_Nx1(lon)
    alt,N3 = input_check_Nx1(alt)
    
    if( (N1!=N2) or (N2!=N3) or (N1!=N3) ):
        raise ValueError('Inputs are not of the same dimension')
    
    if(model=='wgs84'):
        Rew,Rns = earthrad(lat,lat_unit=latlon_unit)
    else:
        Rew = 6378137.0
    
    if(latlon_unit=='deg'):
        lat = np.deg2rad(lat)
        lon = np.deg2rad(lon)
    
    x = (Rew + alt)*np.cos(lat)*np.cos(lon)
    y = (Rew + alt)*np.cos(lat)*np.sin(lon)
    z = ( (1-(6.69437999014e-3))*Rew + alt )*np.sin(lat)
    
    ecef = np.vstack((x,y,z)).T

    if(N1==1):
        ecef = ecef.reshape(3)

    return ecef

def ecef2lla(ecef, latlon_unit='deg'):
    """
    Calculate the Latitude, Longitude and Altitude of a point located on earth 
    given the ECEF Coordinates.
    
    References
    ----------
    .. [1] Jekeli, C.,"Inertial Navigation Systems With Geodetic
       Applications", Walter de Gruyter, New York, 2001, pp. 24
    
    Parameters
    ----------
    ecef : {(N,3)} array like input of ECEF coordinate in X, Y, and Z column, unit is meters
    latlon_unit : {('deg','rad')} specifies the output latitude and longitude unit
    
    Returns
    -------
    lat : {(N,)} array like latitude in unit specified by latlon_unit
    lon : {(N,)} array like longitude in unit specified by latlon_unit
    alt : {(N,)} array like altitude in meters
    """
    ecef,N = input_check_Nx3(ecef)
    if(N>1):
        x = ecef[:,0] 
        y = ecef[:,1]
        z = ecef[:,2]
    else:
        x = ecef[0,0]
        y = ecef[0,1]
        z = ecef[0,2]

    lon = np.arctan2(y,x)

    # Iteration to get Latitude and Altitude
    p = np.sqrt(x**2 + y**2)
    lat = np.arctan2(z,p*(1-(6.69437999014e-3)))

    err = np.ones(N)

    while(np.max(np.abs(err))>1e-10):
        Rew,Rns = earthrad(lat,lat_unit='rad')
        h = p/np.cos(lat) - Rew
        
        err = np.arctan2(z*(1+(6.69437999014e-3)*Rew*np.sin(lat)/z),p) - lat
        
        lat = lat + err
    
    if(latlon_unit=='deg'):
        lat = np.rad2deg(lat)
        lon = np.rad2deg(lon)

    return lat, lon, h

def ned2lla(ned, lat_ref, lon_ref, alt_ref, latlon_unit='deg', alt_unit='m', model='wgs84'):
    """
    Calculate the Latitude, Longitude and Altitude of points given by NED coordinates
    where NED origin given by lat_ref, lon_ref, and alt_ref.
    Parameters
    ----------
    ned : {(N,3)} array like input of NED coordinate in N, E, and D column, unit is meters
    lat_ref : Reference latitude, unit specified by latlon_unit, default in deg
    lon_ref : Reference longitude, unit specified by latlon_unit, default in deg
    alt_ref : Reference altitude, unit specified by alt_unit, default in m
    latlon_unit : {('deg','rad')} specifies the output latitude and longitude unit
    
    Returns
    -------
    lat : {(N,)} array like latitude in unit specified by latlon_unit
    lon : {(N,)} array like longitude in unit specified by latlon_unit
    alt : {(N,)} array like altitude in meters
    Note
    ----
    This method is a wrapper on ned2ecef (add ecef of NED-origin) and ecef2lla.
    """

    ecef = ned2ecef(ned,lat_ref,lon_ref,alt_ref,latlon_unit=latlon_unit,
                                                alt_unit=alt_unit,
                                                model=model)
    # Add vector to ecef representation of NED-origin
    ecef_ref = lla2ecef(lat_ref, lon_ref, alt_ref, latlon_unit=latlon_unit,
                                                   alt_unit=alt_unit,
                                                   model=model)
    ecef += ecef_ref

    lla = ecef2lla(ecef, latlon_unit=latlon_unit)

    return lla

def input_check_Nx1(x):
    """
    Check x to be of dimension Nx1 and reshape it as a 1-D array
    Adhika Lie
    """
    x = np.atleast_1d(x)
    theSize = np.shape(x)

    if(len(theSize) > 1):
        # 1. Input must be of size N x 1
        if ((theSize[0] != 1) & (theSize[1] != 1)):
            raise ValueError('Not an N x 1 array')
        # 2. Make it into a 1-D array
        x = x.reshape(np.size(x))
    elif (theSize[0] == 1):
        x = x[0]

    return x, np.size(x)


def input_check_Nx3(x):
    """
    Check x to be of dimension Nx3
    Adhika Lie
    """
    x = np.atleast_2d(x)
    theSize = np.shape(x)

    if(len(theSize) > 1):
        # 1. Input must be of size N x 3
        if ((theSize[0] != 3) & (theSize[1] != 3)):
            raise ValueError('Not a N x 3 array')
        # 2. Make it into a Nx3 array
        if (theSize[1] != 3):
            x = x.T
        N = x.shape[0]
        # 3. If N == 1, make it into a 1-D array
        if (x.shape[0] == 1):
            x = x.reshape(x.shape[1])

    return x, N


def input_check_Nx3x3(x):
    """
    Check x to be of dimension Nx3x3
    Jacob Niehus
    """
    theSize = np.shape(x)
    N = 1

    if(len(theSize) > 2):
        # 1. Input must be of size N x 3
        if (3, 3) not in (theSize[:2], theSize[-2:]):
            raise ValueError('Not a N x 3 x 3 array')
        # 2. Make it into a Nx3x3 array
        if (theSize[1:] != (3, 3)):
            x = np.rollaxis(x, -1)
        N = x.shape[0]
        # 3. If N == 2, make it into a 2-D array
        if (x.shape[0] == 1):
            x = x[0]
    elif(theSize != (3, 3)):
        raise ValueError('Not a 3 x 3 array')

    return x, N

def earthrad(lat, lat_unit='deg', model='wgs84'):
    """
    Calculate radius of curvature in the prime vertical (East-West) and 
    meridian (North-South) at a given latitude.
    Parameters
    ----------
    lat : {(N,)} array like latitude, unit specified by lat_unit, default in deg
    
    Returns
    -------
    R_N : {(N,)} array like, radius of curvature in the prime vertical (East-West)
    R_M : {(N,)} array like, radius of curvature in the meridian (North-South)
    
    Examples
    --------
    >>> import numpy as np
    >>> from navpy import earthrad
    >>> lat = 0
    >>> Rtransverse, Rmeridian = earthrad(lat)
    >>> Rtransverse
    6378137.0
    >>> Rmeridian
    6335439.3272928288
    >>> lat = [0, np.pi/2]
    >>> Rtransverse, Rmeridian = earthrad(lat,lat_unit='rad')
    >>> Rtransverse
    array([ 6378137.        ,  6399593.62575849])
    >>> Rmeridian
    array([ 6335439.32729283,  6399593.62575849])
    """
    if(lat_unit=='deg'):
        lat = np.deg2rad(lat)
    elif(lat_unit=='rad'):
        pass
    else:
        raise ValueError('Input unit unknown')

    if(model=='wgs84'):
        R_N = 6378137.0/(1-(6.69437999014e-3)*np.sin(lat)**2)**0.5
        R_M = (6378137.0)*(1-(6.69437999014e-3))/(1-(6.69437999014e-3)*np.sin(lat)**2)**1.5
    else:
        raise ValueError('Model unknown')
    
    return R_N, R_M
