#!/usr/bin/python
import rospy
import time
import numpy as np
from numpy import matrix
from numpy import pi, nan
from numpy.matlib import eye, ones, zeros, diag
from numpy.linalg import norm
from numpy import shape
from ekfLib import *
from scipy.linalg import block_diag as blkdiag
import math
from std_msgs.msg import Float64, UInt8
from std_msgs.msg import Bool
from custom_msgs.msg import NorthEastHeading, ImuSensor, gpsFIX,gnssGGA, gnssHDT, gnssRMC
from custom_msgs.msg import twistEstimate, ekfBias, EKF_Kmat, EKF_Pmat, residualPose, residualIMU,inertialSensor
from custom_msgs.srv import ChangeToDP
from geometry_msgs.msg import Point, Vector3, Quaternion
from dynamic_reconfigure.parameter_generator_catkin import *
from dynamic_reconfigure.server import Server
from dp_controller.cfg import resetnedConfig


class EKF:

    def __init__(self):
        print 'EKF initiated'
        self.parameters = self.parameters()
        self.resetFlag = 0
        self.gnssMeasurements = np.zeros((3,3))
        self.imuMeasurements = np.zeros((6,3))
        self.PoseResiduals = np.zeros((3,3))
        self.initialized = 0
    class parameters:

        def __init__(self):
            self.zG = np.mat([0.0, 0.0, 9.81])
            # Known bias
            self.gravityE = np.mat([0, 0, 0]).T 
            # Observation Matrix
        
            gnss = np.bmat([np.eye(3), np.zeros((3, 13))])
            attitude = np.bmat([np.zeros((4,3)),np.eye(4),np.zeros((4,9))])
            velo_imu = np.bmat([np.zeros((3,7)),np.eye(3),np.zeros((3,6))])
            velo_gnss = np.bmat([np.zeros((2,7)),np.eye(2),np.zeros((2,7))])
            self.H = np.bmat([[gnss], [attitude],[velo_imu],[velo_gnss]])
            
            
            # Standard Deviations
            std_accel = 0.25 
            std_gyros = 0.05
            std_pose_GNSS = 1.2
            std_velo_GNSS = 0.15
            
            # Measurement Covariances
            self.R_PoseGNSS  = np.eye(3)*std_pose_GNSS**2
            self.R_VeloGNSS = np.eye(2)*std_velo_GNSS**2
            self.R_Gyros = np.eye(4)*std_gyros**2
            self.R_Accel = 0*np.eye(3)*std_accel**2
            
            # Measurement Covariance Matrix
            self.R = blkdiag(self.R_PoseGNSS, self.R_Gyros,self.R_Accel,self.R_VeloGNSS)

            # Process Covariances 
            qPosition = (0.455*np.eye(3))**2
            qAttitude = (0.01*np.eye(4))**2
            qVelocity = (0.22*np.eye(3))**2
            qAccelBias = np.eye(3)**2
            qGyroBias = np.eye(3)**2
            # Process Covariance Matrix 
            self.Q  = blkdiag(qPosition,qAttitude,qVelocity,qAccelBias,qGyroBias) 


    def NoPrevStateNorState(self, currentTime, filterAttitude):
        self.deltaT = 0 
        if hasattr(self, 'zG') and hasattr(self, 'zC'):
            filterAttitude[2] = self.zC
            q = ypr2quat(filterAttitude)
            self.prevstates = matrix([
                float(self.zG[0]),
                float(self.zG[1]),
                float(self.zG[2]),
                np.asscalar(q[0]),
                np.asscalar(q[1]),
                np.asscalar(q[2]),
                np.asscalar(q[3]),
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                ]).T
                

            self.states = matrix([
                float(self.zG[0]),
                float(self.zG[1]),
                float(self.zG[2]),
                np.asscalar(q[0]),
                np.asscalar(q[1]),
                np.asscalar(q[2]),
                np.asscalar(q[3]),
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                ]).T
            self.velocityBody = np.zeros((3,1))
            self.P_k = np.eye(16)
            self.P_k[1,1] = 0.627
            self.P_k[2,2] = 0.627
            self.P_k[3,3] = 0.63
            self.P_k[4,4] = 0.001
            self.P_k[5,5] = 0.001
            self.P_k[6,6] = 0.001
            self.P_k[7,7] = 0.001
            self.P_k[8,8] = 0.0337
            self.P_k[9,9] = 0.0337
            self.P_k[10,10] = 0.1345




            self.Innovation = zeros((16,1))
            self.kalmanGain = np.zeros((16,10))
            print("State and prev state reset")
            self.prevTime = currentTime


    def ResetEKFObserver(self, currentTime, imuData, filterAttitude): 
        EKFObserverReset = 0
        i = 0
        while EKFObserverReset < 0.5 and i < 20:
            if hasattr(self, 'zG') and hasattr(self, 'zC'):
                self.NoPrevStateNorState(currentTime, filterAttitude)
                self.resetFlag = 1
                EKFObserverReset = 1
                print('EKFObserver successfully reset. (residual)')
            else:
                i = i+1
                time.sleep(0.1)

        if EKFObserverReset < 0.5:
            print('EKFObserver (residual) not reset. Please activate DP mode once more')

    def compassMea(self, zC): 
        if not np.isnan(zC):
            self.zC = zC
        else:
            if hasattr(self, 'zC'):
                del self.zC
                self.zC = zC

    def gpsMea(self, zG):  
        if not np.isnan(zG[0]) or not np.isnan(zG[1]) or not np.isnan(zG[2]):
            self.zG = zG
        else:
            if hasattr(self, 'zG'):
                del self.zG
                self.zG = zG

    def sogcogMea(self, sog,cog):
        if not (np.isnan(sog) and np.isnan(cog)):
            self.zS = sog
            self.zCG = cog*(np.pi/180)
        else:
            if hasattr(self, 'zS'):
                del self.zS
                del self.zCG
    def iterate(self, currentTime, imuData, filterAttitude, gnssData={}):
        if not hasattr(self, 'prevstates') or not hasattr(self, 'prevTime'):
            self.NoPrevStateNorState(currentTime, filterAttitude)
            print 'init state (residual)'
            return
        param = self.parameters        
        # Calculate time passed since last iteration

        deltaT = currentTime - self.prevTime
        self.prevTime = currentTime
        self.deltaT = deltaT
        
        statesEst_pre = self.states
        

        if self.resetFlag == 1:
            print("Reset at start checkpoint")
            return
        # Compute Rotation matrix based on last quaternion measurement
        q_states = statesEst_pre[3:7]
        q_states_normalized = quatNormalizer(q_states)
        RotationQuat = rotationmatrix_quat(q_states_normalized)


        # Extract Measurements         
        LinearAcceleration_m = imuData[0:3]
        AngularVelocity_m = imuData[3:6]
        Velocity_m_imu = RotationQuat.dot(imuData[6:9])
        quaternion_m_imu = ypr2quat(filterAttitude)
        #Normalize
        quaternion_m_imu = quatNormalizer(quaternion_m_imu)
        
        if hasattr(self, 'zG'):
            posMeasuredN = self.zG[0]
            posMeasuredE = self.zG[1]
            posMeasuredD = self.zG[2]
            VelocityN_m_GNSS = np.matrix([self.zS*np.cos(self.zCG)])
            VelocityE_m_GNSS = np.matrix([self.zS*np.sin(self.zCG)])
            measurements = np.bmat([[posMeasuredN],[posMeasuredE],[posMeasuredD],[quaternion_m_imu],[Velocity_m_imu],[VelocityN_m_GNSS],[VelocityE_m_GNSS]])
        else:
            posMeasuredN = statesEst_pre[0]
            posMeasuredE = statesEst_pre[1]
            posMeasuredD = statesEst_pre[2]
            rospy.logwarn("No GNSS measurement")


        # Prediction Step 
        
        position_dot = statesEst_pre[7:10]
        position_bar = statesEst_pre[0:3] + position_dot*deltaT

        
        quaternion_dot = quaternionDot(statesEst_pre[3:7],(AngularVelocity_m-statesEst_pre[13:16]))
        self.qdot = quaternion_dot
        quaternion_bar = statesEst_pre[3:7] + quaternion_dot*deltaT #quatMultiply(statesEst_pre[3:7],quaternion_dot*deltaT) #quatMultiply(quaternion_dot*deltaT,statesEst_pre[3:7])
        #Normalize quaternion
        quaternion_bar = quatNormalizer(quaternion_bar)


        velo_dot = (RotationQuat.dot(LinearAcceleration_m - statesEst_pre[10:13]) + param.gravityE)
        velo_bar = statesEst_pre[7:10] + velo_dot*deltaT


        if self.resetFlag == 1:
            self.NoPrevStateNorState(currentTime, filterAttitude)
            print("Reset at second checkpoint")
            return

        ## State Propagation Matrix f(X,u)
        fX = np.bmat([[position_bar],[quaternion_bar],[velo_bar],[np.zeros((3,1))],[np.zeros((3,1))]])
        
        # Jacobian of State Propagation Matrix => d(fX)/dX
        FX_jacobi = statePropagJacobian(deltaT,imuData[3:6],statesEst_pre[3:7],imuData[0:3],statesEst_pre[10:13],statesEst_pre[13:16])
        
        P_k = self.P_k
        P_bar_kPlus = np.linalg.multi_dot([FX_jacobi,P_k,FX_jacobi.T]) + param.Q

       
        if self.resetFlag == 1:
            self.NoPrevStateNorState(currentTime, filterAttitude)
            print("Reset at third checkpoint")
            return        
        
        
        # Predicted States
        
        stateBar = fX

        ## UPDATE 
        ## Residual Covariance => H*P*H^T + R
        
        S = np.linalg.multi_dot([param.H,P_bar_kPlus,param.H.T]) + param.R
        
        ## Kalman Gain Matrix => P*H^T / H*P*H^T + R
        
        K = np.linalg.multi_dot([P_bar_kPlus,param.H.T,S.I])
        self.kalmanGain = K
        ## Extract residuals
        self.nedEstimate = (param.H.dot(stateBar))[0:3]

        ## State Estimate Update
        
        # Quaternion State Estimate Update (Own procedure due to preservation of the unit quaternion)
        
        quatInnovationTerm = quatInnovation(measurements[3:7],stateBar[3:7])
        self.attitudeResidual = quatInnovationTerm
        quatHat = quatStateUpdate(stateBar[3:7],K[4:7,4:7],quatInnovationTerm)
        
        # Regular State Estimation Update (all states except quaternion)
        self.Innovation = measurements-param.H.dot(stateBar)
        
        
        stateHat  = stateBar + K*(self.Innovation)
        # Overwrite non-normalized quaternion with the quatHat 
        stateHat[3:7] = quatHat
        
        
        ## Updated State Covariance 
        
        P_kPlus = P_bar_kPlus - np.linalg.multi_dot([K,S,K.T])
        
        if self.resetFlag == 1:
            del self.kalmanGain
            del self.Innovation
            self.NoPrevStateNorState(currentTime, filterAttitude)
            print("Reset at last checkpoint")
            return
            
        ## Advance to the next step
        
        self.P_k = P_kPlus
        self.states = stateHat
        
        
 

class Observer:

    def __init__(self):
        print 'Observer initiated'
        self.lat0 = get_param('~lat0', 63.439175163)
        self.lon0 = get_param('~lon0', 10.400329186)
        self.alt0 = get_param('~alt0', -0.682)

        self.ekf = EKF()        
        
        # Init publishers
        self.readySignalPub = rospy.Publisher("residualobserver/resetNED", UInt8, queue_size=10)
        self.residualsPose = rospy.Publisher('residualobserver/residualsPose', residualPose, queue_size=10)
        self.residualsAttitude = rospy.Publisher('residualobserver/residualsAttitude',Point, queue_size=10)
        self.residualsVelocity = rospy.Publisher('residualobserver/residualsVelocity',Point, queue_size=10)
        self.residualsIMU = rospy.Publisher("residualobserver/residualsIMU",residualIMU, queue_size=10)
        self.observedFixPub = rospy.Publisher('residualobserver/fix', Point, queue_size=10)
        self.pub_eta_GNSS = rospy.Publisher('residualobserver/eta_GNSS', NorthEastHeading, queue_size=1)
        self.dropped = 0

        # Init callbacks
        FIXss1 = rospy.Subscriber("/signal/GNSS_pre1", gnssGGA, self.FIXcb)
        FIXss2 = rospy.Subscriber("/signal/GNSS_pre2", gnssGGA, self.FIXcb)
        FIXss3 = rospy.Subscriber("/signal/GNSS_pre3", gnssGGA, self.FIXcb)
        IMUss1 = rospy.Subscriber("/signal/IMU_pre1", inertialSensor, self.IMUcb_actual)
        IMUss2 = rospy.Subscriber("/signal/IMU_pre2", inertialSensor, self.IMUcb_actual)
        IMUss3 = rospy.Subscriber("/signal/IMU_pre3", inertialSensor, self.IMUcb_actual)
        IMUss = rospy.Subscriber("xsens/imu", ImuSensor, self.IMUcb_sim)
        HDTss = rospy.Subscriber("vectorVS330/heading", gnssHDT, self.HDTcb)
        VELss = rospy.Subscriber("vectorVS330/velocity", gnssRMC, self.SOGCOGcb)        
        DPss = rospy.Service("change_to_DP_residual", ChangeToDP, self.parameter_callback_changeToDP)
        rospy.loginfo("Initialized")






    def FIXcb(self, data):

        self.latitude = data.latitude
        self.longitude = data.longitude
        self.altitude = data.altitude
        self.unit = data.gnss_unit

        # Insert measurements in GNSS measurement matrix, used for pose residual generation:
        if(data.gnss_unit == 1):
            self.ekf.gnssMeasurements[0,0] = data.latitude
            self.ekf.gnssMeasurements[1,0] = data.longitude
            self.ekf.gnssMeasurements[2,0] = data.altitude + 0.22

        elif(data.gnss_unit == 2):
            self.ekf.gnssMeasurements[0,1] = data.latitude
            self.ekf.gnssMeasurements[1,1] = data.longitude
            self.ekf.gnssMeasurements[2,1] = data.altitude + 0.22

        elif(data.gnss_unit == 3):
            self.ekf.gnssMeasurements[0,2] = data.latitude
            self.ekf.gnssMeasurements[1,2] = data.longitude
            self.ekf.gnssMeasurements[2,2] = data.altitude + 0.22
        
        
        
        if np.isnan(self.latitude) or np.isnan(self.longitude) or np.isnan(self.altitude):
            rospy.logwarn('No GPS measurements for observer')
            return
        if not hasattr(self, 'heading') or np.isnan(self.heading):
            rospy.logwarn('No compass measurements for observer')
            return
        if not hasattr(self, 'roll') or not hasattr(self, 'pitch') or np.isnan(self.roll) or np.isnan(self.pitch):
            rospy.logwarn('No attitude measurements for observer')
            return
        self.NED = lla2ned(self.latitude, self.longitude, self.altitude, self.lat0, self.lon0, self.alt0)

        # Do lever arm comp
        self.NED = levercomp(self.NED, self.heading, self.roll, self.pitch)
        self.ekf.gpsMea(self.NED)

        # Publish
        eta = NorthEastHeading()
        eta.pos_heading = self.heading * (180 / pi)
        eta.pos_north = self.NED[0]
        eta.pos_east = self.NED[1]
        self.pub_eta_GNSS.publish(eta)
        if(hasattr(self.ekf,'nedEstimate')):
            lla_estimate = ned2lla(self.ekf.nedEstimate.T, self.lat0, self.lon0, self.alt0, latlon_unit='deg', alt_unit='m', model='wgs84')
            pose_estimate = np.array([lla_estimate[0],lla_estimate[1],lla_estimate[2]])
            pose_estimate = pose_estimate.reshape(3,1)
            pose_estimate = np.hstack([pose_estimate,pose_estimate,pose_estimate])
            self.ekf.PoseResiduals = self.ekf.gnssMeasurements - pose_estimate

        
    
    ''' Observer subscribes to Speed Over Ground, Course Over ground (no filtering atm) '''
    def SOGCOGcb(self, data):
        self.sog = data.speed_mps
        #rospy.logwarn("SOG: %f COG: %f",data.speed_mps, data.course)
        if(data.course > 180.0):
            self.cog = data.course - 360
        else:
            self.cog = data.course
        
        if np.isnan(self.sog) or np.isnan(self.cog):
            rospy.logwarn('no sog/cog measurment')
            return
        self.ekf.sogcogMea(self.sog,self.cog)




    def HDTcb(self, data):  # Yaw angle from GPHDT is in degrees and wrapped
        self.heading = data.heading * (np.pi / 180)
        #self.heading = wrapHeading_rad(self.heading)
        self.ekf.compassMea(self.heading)
    
    def IMUcb_actual(self, data):
        if(data.imuUnit == 1):
            self.ekf.imuMeasurements[0,0] = data.roll
            self.ekf.imuMeasurements[1,0] = data.pitch
            self.ekf.imuMeasurements[2,0] = data.yaw
            self.ekf.imuMeasurements[3,0] = data.rollRate
            self.ekf.imuMeasurements[4,0] = data.pitchRate
            self.ekf.imuMeasurements[5,0] = data.yawRate
        if(data.imuUnit == 2):
            self.ekf.imuMeasurements[0,1] = data.roll
            self.ekf.imuMeasurements[1,1] = data.pitch
            self.ekf.imuMeasurements[2,1] = data.yaw
            self.ekf.imuMeasurements[3,1] = data.rollRate
            self.ekf.imuMeasurements[4,1] = data.pitchRate
            self.ekf.imuMeasurements[5,1] = data.yawRate
        if(data.imuUnit == 3):
            self.ekf.imuMeasurements[0,2] = data.roll
            self.ekf.imuMeasurements[1,2] = data.pitch
            self.ekf.imuMeasurements[2,2] = data.yaw
            self.ekf.imuMeasurements[3,2] = data.rollRate
            self.ekf.imuMeasurements[4,2] = data.pitchRate
            self.ekf.imuMeasurements[5,2] = data.yawRate
            
    def IMUcb_sim(self, data):
        if not hasattr(self, 'heading') or np.isnan(self.heading):
            rospy.logwarn('No compass measurements for observer')
            return

        self.roll = data.orientation.roll * (np.pi / 180)
        self.pitch = data.orientation.pitch * (np.pi / 180)
        yaw = data.orientation.yaw * (np.pi / 180)
        self.imuData = np.array([[data.linear_acceleration.x],
                                 [data.linear_acceleration.y],
                                 [data.linear_acceleration.z],
                                 [data.angular_velocity.x * (np.pi / 180)],
                                 [data.angular_velocity.y * (np.pi / 180)],
                                 [data.angular_velocity.z * (np.pi / 180)],
                                 [data.linear_velocity.x],
                                 [data.linear_velocity.y],
                                 [data.linear_velocity.z]])
        self.filterAttitude = matrix([self.roll,
                                      self.pitch, yaw]).T
        self.timestamp = rospy.Time.now().secs+rospy.Time.now().nsecs * 10 ** (-9)

        self.ekf.iterate(self.timestamp, self.imuData,self.filterAttitude)
        if self.ekf.resetFlag == 1:
            self.ekf.NoPrevStateNorState(self.timestamp, self.filterAttitude)
            self.ekf.resetFlag = 0
            return
        if hasattr(self.ekf, 'states') and hasattr(self.ekf, 'attitudeResidual') and hasattr(self.ekf, 'PoseResiduals'):
            ## Publish PVA residuals (y_measured - x_hat)
            for i in range(0,3):
                rPose = residualPose()
                rPose.x = self.ekf.PoseResiduals[0,i]
                rPose.y = self.ekf.PoseResiduals[1,i]
                rPose.z = self.ekf.PoseResiduals[2,i]
                rPose.unit = i + 1
                if(self.ekf.initialized == 0):
                    readySignal = UInt8()
                    readySignal.data = 1
                    self.readySignalPub.publish(readySignal)
                    self.ekf.initialized = 1
                self.residualsPose.publish(rPose)
            eulAngle = quat2ypr(self.ekf.states[3:7]) * (180 / pi)
            q = self.ekf.states[3:7]
            # Normalize
            q = quatNormalizer(q)
            
            R = rotationMatrixEuler(eulAngle[0] * pi / 180, eulAngle[1] * pi / 180, eulAngle[2] * pi / 180)
            T = angularVelTransformation_quat(q)
            q = self.ekf.states[3:7]
            # Normalize
            q = quatNormalizer(q)
            angularVelBody = T.T.dot(quatDifferentialVik(q, (self.imuData[3:6] - self.ekf.states[13:16])))*231 # The scaling factor has been found by studying actual yaw rate values and those produced by QuatDifferential
            # message.roll,message.pitch,message.yaw,message.rollRate,message.pitchRate,message.yawRate
            for i in range(0,3):
                rIMU = residualIMU()
                rIMU.roll = self.ekf.imuMeasurements[0,i] - eulAngle[0]
                rIMU.pitch = self.ekf.imuMeasurements[1,i] - eulAngle[1]
                rIMU.yaw = data.orientation.yaw - eulAngle[2] 
                rIMU.rollRate = self.ekf.imuMeasurements[3,i] - angularVelBody[0]
                rIMU.pitchRate = self.ekf.imuMeasurements[4,i] - angularVelBody[1]
                rIMU.yawRate = data.angular_velocity.z- angularVelBody[2] 
                rIMU.unit = i + 1
                self.residualsIMU.publish(rIMU)

            rAttitude = Point()
            q_residual = self.ekf.attitudeResidual
            rAttitude.x = q_residual[0]
            rAttitude.y = q_residual[1]
            rAttitude.z = q_residual[2]
            rVelocity = Point()
            rVelocity.x = self.ekf.Innovation[7]
            rVelocity.y = self.ekf.Innovation[8]
            rVelocity.z = self.ekf.Innovation[9]
            fix = Point()
            fix.x = self.ekf.states[0]
            fix.y = self.ekf.states[1]
            fix.z = self.ekf.states[2]
            self.observedFixPub.publish(fix)
            self.residualsAttitude.publish(rAttitude)
            self.residualsVelocity.publish(rVelocity)



    def parameter_callback(self, mode, level):
        if mode.set == 1:
            rospy.set_param('~lon0', self.longitude)
            rospy.set_param('~lat0', self.latitude)
            rospy.set_param('~alt0', self.altitude)
            print self.longitude
            print self.latitude
            print self.altitude
            self.lat0 = get_param('~lat0', 63.439206)
            self.lon0 = get_param('~lon0', 10.399675)
            self.alt0 = get_param('~alt0', 39)

            self.NED = lla2ned(self.latitude, self.longitude, self.altitude, self.lat0, self.lon0, self.alt0)
            mode.set = 0
        return mode
    
    def parameter_callback_changeToDP(self,level):
        rospy.set_param('~lon0', self.longitude)
        rospy.set_param('~lat0', self.latitude)
        rospy.set_param('~alt0', self.altitude)
        print self.longitude
        print self.latitude
        print self.altitude
        self.lat0 = get_param('~lat0', 63.439206)
        self.lon0 = get_param('~lon0', 10.399675)
        self.alt0 = get_param('~alt0', 39)
        self.NED = lla2ned(self.latitude, self.longitude, self.altitude, self.lat0, self.lon0, self.alt0)
        print('Reseting EKF residual observer')
        self.ekf.ResetEKFObserver(self.timestamp, self.imuData ,self.filterAttitude)
        return True

def get_param(name, default):
    try:
        v = rospy.get_param(name)
        rospy.loginfo("Found parameter: %s, value: %s" % (name, str(v)))
    except KeyError:
        v = default
        rospy.logwarn("Cannot find value for parameter: %s, assigning "
                      "default: %s" % (name, str(v)))
    return v


# Initiate
def main():

    # initiate node
    rospy.init_node('ResidualObserverNode')
    observer = Observer()
    srv = Server(resetnedConfig, observer.parameter_callback)
    rospy.spin()

    # while not rospy.is_shutdown():
    #   main()


if __name__ == '__main__':
    try:
        main()
    except rospy.ROSInterruptException:
        pass
