
#include "testnode.h"


void SigBlock::initPublishers(){
	GNSS_posteriori_pub = nh.advertise<custom_msgs::gnssGGA>("/signal/GNSS_post", 1);
	IMU_posteriori_pub = nh.advertise<custom_msgs::inertialSensor>("/signal/IMU_post", 1);
    gnss_outlier_pub = nh.advertise<custom_msgs::gnssGGA>("/signal/GNSS_check_outlier", 1);
    imu_outlier_pub = nh.advertise<custom_msgs::inertialSensor>("/signal/IMU_check_outlier", 1);

    gnss_residual_pub1 = nh.advertise<custom_msgs::gnssGGA>("/signal/GNSS_pre1", 1);
    gnss_residual_pub2 = nh.advertise<custom_msgs::gnssGGA>("/signal/GNSS_pre2", 1);
    gnss_residual_pub3 = nh.advertise<custom_msgs::gnssGGA>("/signal/GNSS_pre3", 1);

    imu_residual_pub1 = nh.advertise<custom_msgs::inertialSensor>("/signal/IMU_pre1", 1);
    imu_residual_pub2 = nh.advertise<custom_msgs::inertialSensor>("/signal/IMU_pre2", 1);
    imu_residual_pub3 = nh.advertise<custom_msgs::inertialSensor>("/signal/IMU_pre3", 1);
}

void SigBlock::initSubscribers(){
    const int gnssNum_1 = 1;
    const int gnssNum_2 = 2;
    const int gnssNum_3 = 3;
    const int imuNum_1 = 1;
    const int imuNum_2 = 2;
    const int imuNum_3 = 3;
    

    gnssGGA1_sub = nh.subscribe<custom_msgs::gpsFIX>("/virtualGPS1/pre",10,boost::bind(&SigBlock::gnssCB, this, _1, gnssNum_1));
    gnssGGA2_sub = nh.subscribe<custom_msgs::gpsFIX>("/virtualGPS2/pre",10,boost::bind(&SigBlock::gnssCB, this, _1, gnssNum_2));
    gnssGGA3_sub = nh.subscribe<custom_msgs::gpsFIX>("/virtualGPS3/pre",10,boost::bind(&SigBlock::gnssCB, this, _1, gnssNum_3));
    
    IMU1_sub = nh.subscribe<custom_msgs::inertialSensor>("/virtualInertial1/pre",10,boost::bind(&SigBlock::imuCB, this, _1, imuNum_1));
    IMU2_sub = nh.subscribe<custom_msgs::inertialSensor>("/virtualInertial2/pre",10,boost::bind(&SigBlock::imuCB, this, _1, imuNum_2));    
    IMU3_sub = nh.subscribe<custom_msgs::inertialSensor>("/virtualInertial3/pre",10,boost::bind(&SigBlock::imuCB, this, _1, imuNum_3));
    
    gnss_outlierReport_sub = nh.subscribe("/signal/GNSS_outlier_report", 10, &SigBlock::GNSSoutlierCB, this);
    imu_outlierReport_sub = nh.subscribe("/signal/IMU_outlier_report", 10, &SigBlock::IMUoutlierCB, this);
    residual_pose_sub = nh.subscribe("residualobserver/residualsPose", 10, &SigBlock::gnssResidualCB, this);
    residual_imu_sub = nh.subscribe("residualobserver/residualsIMU", 10, &SigBlock::imuResidualCB, this);
    residual_generator_ready_sub = nh.subscribe("residualobserver/resetNED", 10, &SigBlock::residualReadyCB,this);
    residual_generator_ready_IMU_sub = nh.subscribe("residualobserver/resetNED", 10, &SigBlock::IMUresidualReadyCB,this);
}


void SigBlock::gnssCB(const custom_msgs::gpsFIX::ConstPtr& msg, const int GNSS_Num){
    custom_msgs::gpsFIX message = *msg;
	if(std::isnan(message.latitude)){
		ROS_WARN("Error (NaN) in GPS %i latitude measurement!", GNSS_Num);		
		return;
	}
	if(std::isnan(message.longitude)){
		ROS_WARN("Error (NaN) in GPS %i longitude measurement!", GNSS_Num);
		return;
	}
	if(std::isnan(message.altitude)){
		ROS_WARN("Error (NaN) in GPS %i altitude measurement!", GNSS_Num);
		return;	
	}
    GNSS_Data[GNSS_Num-1].latitude =  message.latitude;
    GNSS_Data[GNSS_Num-1].longitude = message.longitude;
    GNSS_Data[GNSS_Num-1].altitude = message.altitude;
    GNSS_Data[GNSS_Num-1].timestamp_double = double(message.timestamp.sec) + double(message.timestamp.nsec)*(1e-9);
    GNSS_Data[GNSS_Num-1].published = false;
    //ROS_WARN("FIRST the double timestamp is %Lf", GNSS_Data[GNSS_Num-1].timestamp_double);		
    ros::Time timeinstance = message.timestamp;
    GNSS_Data[GNSS_Num-1].timestamp = timeinstance;     
    //ROS_WARN("longitude: %f num: %i timestamp: %f", longitude[GNSS_Num-1], GNSS_Num, timestamp_lastUpdate[GNSS_Num-1]);
    gnssFTC();
}

void SigBlock::gnssResidualCB(const custom_msgs::residualPose::ConstPtr& msg){
	custom_msgs::residualPose message = *msg;
	std::vector<double> residuals = {message.x,message.y,message.z};
	std::vector<double> S_values;
	if(GNSS_residualsReady == 1){
		for (auto iter = GNSS_Residual_Data.begin(); iter != GNSS_Residual_Data.end(); ++iter){
			if(message.unit == (*iter).unit){
				(*iter).Vec1.measurement = residuals[0];
				(*iter).Vec2.measurement = residuals[1];
				(*iter).Vec3.measurement = residuals[2];
				
				S_values = {(*iter).Vec1.S_high,(*iter).Vec1.S_low};
				S_values = CUSUM_test((*iter).Vec1.measurement,u_H1_GNSS_horizontal,u_H2_GNSS_horizontal,sigma_null_GNSS_horizontal,sigma_H1_GNSS_horizontal,sigma_H2_GNSS_horizontal,S_values);
				(*iter).Vec1.S_high = S_values[0];
				(*iter).Vec1.S_low = S_values[1];
				
				S_values = {(*iter).Vec2.S_high,(*iter).Vec2.S_low};
				S_values = CUSUM_test((*iter).Vec2.measurement,u_H1_GNSS_horizontal,u_H2_GNSS_horizontal,sigma_null_GNSS_horizontal,sigma_H1_GNSS_horizontal,sigma_H2_GNSS_horizontal,S_values);
				(*iter).Vec2.S_high = S_values[0];
				(*iter).Vec2.S_low = S_values[1];
				
				S_values = {(*iter).Vec3.S_high,(*iter).Vec3.S_low};
				S_values = CUSUM_test((*iter).Vec3.measurement,u_H1_GNSS_vertical,u_H2_GNSS_vertical,sigma_null_GNSS_vertical,sigma_H1_GNSS_vertical,sigma_H2_GNSS_vertical,S_values);
				(*iter).Vec3.S_high = S_values[0];
				(*iter).Vec3.S_low = S_values[1];
			}
		}
	}
}

void SigBlock::imuResidualCB(const custom_msgs::residualIMU::ConstPtr& msg){
	custom_msgs::residualIMU message = *msg;
	std::vector<double> residuals = {message.roll,message.pitch,message.yaw,message.rollRate,message.pitchRate,message.yawRate};
	std::vector<double> S_values;
	if(IMU_residualsReady == 1){
		for (auto iter = IMU_Residual_Data.begin(); iter != IMU_Residual_Data.end(); ++iter){
			if(message.unit == (*iter).unit){
				(*iter).Vec1.measurement = residuals[0];
				(*iter).Vec2.measurement = residuals[1];
				(*iter).Vec3.measurement = residuals[2];
				(*iter).Vec4.measurement = residuals[3];
				(*iter).Vec5.measurement = residuals[4];
				(*iter).Vec6.measurement = residuals[5];

				
				S_values = {(*iter).Vec1.S_high,(*iter).Vec1.S_low};
				S_values = CUSUM_test((*iter).Vec1.measurement,u_H1_IMU,u_H2_IMU,sigma_null_IMU,sigma_H1_IMU,sigma_H2_IMU,S_values);
				(*iter).Vec1.S_high = S_values[0];
				(*iter).Vec1.S_low = S_values[1];
				
				S_values = {(*iter).Vec2.S_high,(*iter).Vec2.S_low};
				S_values = CUSUM_test((*iter).Vec2.measurement,u_H1_IMU,u_H2_IMU,sigma_null_IMU,sigma_H1_IMU,sigma_H2_IMU,S_values);
				(*iter).Vec2.S_high = S_values[0];
				(*iter).Vec2.S_low = S_values[1];
				
				S_values = {(*iter).Vec3.S_high,(*iter).Vec3.S_low};
				S_values = CUSUM_test((*iter).Vec3.measurement,u_H1_IMU,u_H2_IMU,sigma_null_IMU,sigma_H1_IMU,sigma_H2_IMU,S_values);
				(*iter).Vec3.S_high = S_values[0];
				(*iter).Vec3.S_low = S_values[1];
				
				S_values = {(*iter).Vec4.S_high,(*iter).Vec4.S_low};
				S_values = CUSUM_test((*iter).Vec4.measurement,u_H1_IMU,u_H2_IMU,sigma_null_IMU,sigma_H1_IMU,sigma_H2_IMU,S_values);
				(*iter).Vec4.S_high = S_values[0];
				(*iter).Vec4.S_low = S_values[1];

				S_values = {(*iter).Vec5.S_high,(*iter).Vec5.S_low};
				S_values = CUSUM_test((*iter).Vec5.measurement,u_H1_IMU,u_H2_IMU,sigma_null_IMU,sigma_H1_IMU,sigma_H2_IMU,S_values);
				(*iter).Vec5.S_high = S_values[0];
				(*iter).Vec5.S_low = S_values[1];

				S_values = {(*iter).Vec6.S_high,(*iter).Vec6.S_low};
				S_values = CUSUM_test((*iter).Vec6.measurement,u_H1_IMU,u_H2_IMU,sigma_null_IMU,sigma_H1_IMU,sigma_H2_IMU,S_values);
				(*iter).Vec6.S_high = S_values[0];
				(*iter).Vec6.S_low = S_values[1];
			}
		}
	}
}

void SigBlock::imuCB(const custom_msgs::inertialSensor::ConstPtr& msg, const int IMU_Num){
    custom_msgs::inertialSensor message = *msg;
    

	if(std::isnan(message.roll)){
		ROS_WARN("Error (NaN) in IMU %i roll measurement!",IMU_Num);		
		return;
	}
	if(std::isnan(message.pitch)){
		ROS_WARN("Error (NaN) in IMU %i pitch measurement!",IMU_Num);		
		return;
	}
	if(std::isnan(message.yaw)){
		ROS_WARN("Error (NaN) in IMU %i yaw measurement!",IMU_Num);		
		return;
	}
	if(std::isnan(message.rollRate)){
		ROS_WARN("Error (NaN) in IMU %i roll rate measurement!",IMU_Num);		
		return;
	}
	if(std::isnan(message.pitchRate)){
		ROS_WARN("Error (NaN) in IMU %i pitch rate measurement!",IMU_Num);		
		return;
	}
	if(std::isnan(message.yawRate)){
		ROS_WARN("Error (NaN) in IMU %i yaw rate measurement!",IMU_Num);
		return;
	}
    IMU_Data[IMU_Num-1].roll =  message.roll;
    IMU_Data[IMU_Num-1].pitch = message.pitch;
    IMU_Data[IMU_Num-1].yaw = message.yaw; 

    IMU_Data[IMU_Num-1].rollRate = message.rollRate;
    IMU_Data[IMU_Num-1].pitchRate = message.pitchRate;
    IMU_Data[IMU_Num-1].yawRate = message.yawRate;
    IMU_Data[IMU_Num-1].published = false;
    
    IMU_Data[IMU_Num-1].timestamp_double = double(message.timestamp.sec) + double(message.timestamp.nsec)*(1e-9);
    ros::Time timeinstance = message.timestamp;
    IMU_Data[IMU_Num-1].timestamp = timeinstance;

    imuFTC();
}

void SigBlock::GNSSoutlierCB(const custom_msgs::gnssOutlier &msg){
    
    if(msg.outlier == -1){
        int unit = msg.gnss_unit;
        ROS_WARN("GNSS OUTLIER DETECTED - UNIT %d",unit);
    }
}

void SigBlock::IMUoutlierCB(const custom_msgs::imuOutlier &msg){
    
    if(msg.outlier == -1){
        int unit = msg.imu_unit;
        ROS_WARN("IMU OUTLIER DETECTED - UNIT %d",unit);
    }
}

void SigBlock::residualReadyCB(const std_msgs::UInt8 &msg){
	if(msg.data == 1){
		GNSS_residualsReady = 1;
	}
}

void SigBlock::IMUresidualReadyCB(const std_msgs::UInt8 &msg){
	if(msg.data == 1){
		IMU_residualsReady = 1;
	}
}

void SigBlock::gnssPub(std::vector<double> publish){
	//if(std::isnan(publish[0]) == false){
		custom_msgs::gnssGGA post_GNSS;
		post_GNSS.latitude = publish[0];
		post_GNSS.longitude = publish[1];
		post_GNSS.altitude = publish[2];
		GNSS_posteriori_pub.publish(post_GNSS);
	//}
}

void SigBlock::imuPub(std::vector<double> publish){
	if(std::isnan(publish[0]) == false){
		custom_msgs::inertialSensor post_IMU;
		post_IMU.roll = publish[0];
		post_IMU.pitch = publish[1];
		post_IMU.yaw = publish[2];
		post_IMU.rollRate = publish[3];
		post_IMU.pitchRate = publish[4];
		post_IMU.yawRate = publish[5];
		IMU_posteriori_pub.publish(post_IMU);
	}
}


double SigBlock::averageVector(std::vector<double> const& input){
    return std::accumulate(input.begin(), input.end(), 0.)/ input.size();
}

bool SigBlock::doubleEqual(double arg1, double arg2, double epsilon) {
	double abs_1 = fabs(arg1);
	double abs_2 = fabs(arg2);
	double diff = fabs(arg1 - arg2);
	
	
	if (arg1 == arg2) { // Best case scenario, also tolerates infinities
		return true;
	} else if (arg1 == 0 || arg2 == 0 || diff < DBL_MIN) {
		// Non-relative errors 
		return diff < (epsilon * DBL_MIN);
	} else { // Relative error check. 
		return (diff/std::min((abs_1 + abs_2), DBL_MAX) < epsilon);
	}
}

void SigBlock::initGNSSContainers(){
    // Initing GNSS data vectors, containing coordinates, timestamps etc. 
    GNSS_CONTAINER DATA;
    DATA.latitude = nan("");
    DATA.longitude = nan("");
    DATA.altitude = nan("");
    DATA.timestamp_double = nan("");
    DATA.deactivated = false;
    DATA.published = true;
    GNSS_Data.assign(NUM_GNSS_UNITS,DATA);
    
    
    // Initing GNSS residual container 
    
    for(int i = 1; i < (NUM_GNSS_UNITS + 1); i++){
		GNSS_RESIDUAL_CONTAINER DATA;
		DATA.Vec1.measurement = nan("");
		DATA.Vec1.S_high = 0;
		DATA.Vec1.S_low = 0;
		DATA.Vec2.measurement = nan("");
		DATA.Vec2.S_high = 0;
		DATA.Vec2.S_low = 0;
		DATA.Vec3.measurement = nan("");
		DATA.Vec3.S_high = 0;
		DATA.Vec3.S_low = 0;
		DATA.unit = i;
		GNSS_Residual_Data.insert(GNSS_Residual_Data.end(),DATA);
	}
    
    // Initing Signal Fault vectors 
    SIG_INFO GNSS_UNIT;
    
    GNSS_UNIT.LostSignal = false;
    GNSS_UNIT.BiasedSignal = false;
    GNSS_UNIT.FrozenSignal = false;
    GNSS_UNIT.HighVarSignal = false;
    GNSS_UNIT.DriftingSignal = false;
    GNSS_UNIT.OutlierAlarm = false;
    GNSS_UNIT.ScalingError = false;
    GNSS_UNIT.InvalidRange = false;
    GNSS_UNIT.DeviationVoting = false;
    GNNS_SignalHealth.assign(NUM_GNSS_UNITS,GNSS_UNIT);
}

void SigBlock::initIMUContainers(){
	
    // Initing IMU data vectors, containing rads and rad/s, timestamps etc. 
    IMU_CONTAINER DATA;
    DATA.roll = nan("");
    DATA.pitch = nan("");
    DATA.yaw = nan("");
    DATA.rollRate = nan("");
    DATA.pitchRate = nan("");
    DATA.yawRate = nan("");
    DATA.timestamp_double = nan("");
    DATA.deactivated = false;
    DATA.published = true;
    IMU_Data.assign(NUM_IMU_UNITS,DATA);


	// Initing IMU residual container 
    
    for(int i = 1; i < (NUM_IMU_UNITS + 1); i++){
		IMU_RESIDUAL_CONTAINER DATA;
		DATA.Vec1.measurement = nan("");
		DATA.Vec1.S_high = 0;
		DATA.Vec1.S_low = 0;
		DATA.Vec2.measurement = nan("");
		DATA.Vec2.S_high = 0;
		DATA.Vec2.S_low = 0;
		DATA.Vec3.measurement = nan("");
		DATA.Vec3.S_high = 0;
		DATA.Vec3.S_low = 0;
		DATA.Vec4.measurement = nan("");
		DATA.Vec4.S_high = 0;
		DATA.Vec4.S_low = 0;
		DATA.Vec5.measurement = nan("");
		DATA.Vec5.S_high = 0;
		DATA.Vec5.S_low = 0;
		DATA.Vec6.measurement = nan("");
		DATA.Vec6.S_high = 0;
		DATA.Vec6.S_low = 0;
		DATA.unit = i;
		IMU_Residual_Data.insert(IMU_Residual_Data.end(),DATA);
	}
	
    // Initing Signal Fault vectors 
    SIG_INFO IMU_UNIT;
    
    IMU_UNIT.LostSignal = false;
    IMU_UNIT.BiasedSignal = false;
    IMU_UNIT.FrozenSignal = false;
    IMU_UNIT.HighVarSignal = false;
    IMU_UNIT.DriftingSignal = false;
    IMU_UNIT.OutlierAlarm = false;
    IMU_UNIT.ScalingError = false;
    IMU_UNIT.InvalidRange = false;
    IMU_UNIT.DeviationVoting = false;
    IMU_SignalHealth.assign(NUM_IMU_UNITS,IMU_UNIT);
}


SigBlock::SigBlock(ros::NodeHandle nh) : nh(nh)
{
    initPublishers();
    initSubscribers();
    initGNSSContainers();
    initIMUContainers();

}
