#include "testnode.h"



// Main Signal Processing Block

void SigBlock::gnssFTC(){

	for(int i = 0; i < GNSS_Data.size(); i++){
		if(std::isnan(GNSS_Data[i].timestamp_double)){
			break;
		}
		if(i == (GNSS_Data.size() - 1)){
			timestampChecker(GNSS_Data, 0);
		}
	}
	
	
	// Publish message for outlier detection mechanism
	for(int i = 0; i < GNSS_Data.size(); i++){
		if((std::isnan(GNSS_Data[i].latitude) == false) && (std::isnan(GNSS_Data[i].longitude) == false) && (std::isnan(GNSS_Data[i].altitude) == false)){
			custom_msgs::gnssGGA outlier_check;
			outlier_check.latitude = GNSS_Data[i].latitude;
			outlier_check.longitude = GNSS_Data[i].longitude;
			outlier_check.altitude = GNSS_Data[i].altitude;
			outlier_check.timestamp = GNSS_Data[i].timestamp;
			outlier_check.gnss_unit = i + 1;
			gnss_outlier_pub.publish(outlier_check);
		}
    }
    
    
	// Send GNSS message to residual generating EKF, which CUSUM will use later
	for(int i = 0; i < GNSS_Data.size(); i++){
		if((std::isnan(GNSS_Data[i].latitude) == false) && (std::isnan(GNSS_Data[i].longitude) == false) && (std::isnan(GNSS_Data[i].altitude) == false) && (GNSS_Data[i].published == false)){
			custom_msgs::gnssGGA residual_calculation;
			residual_calculation.latitude = GNSS_Data[i].latitude;
			residual_calculation.longitude = GNSS_Data[i].longitude;
			residual_calculation.altitude = GNSS_Data[i].altitude;
			residual_calculation.timestamp = GNSS_Data[i].timestamp;
			residual_calculation.gnss_unit = i + 1;
			switch(i){
				case 0: {gnss_residual_pub1.publish(residual_calculation); break;}
				case 1: {gnss_residual_pub2.publish(residual_calculation); break;}
				case 2: {gnss_residual_pub3.publish(residual_calculation); break;}
			}
			GNSS_Data[i].published = true;
		}
   }
   
	int operational_GNSS_units = 0;
	for(int i = 0; i < GNSS_Data.size(); i++){
		if(GNSS_Data[i].deactivated == false && (std::isnan(GNSS_Data[i].latitude) == false) && (std::isnan(GNSS_Data[i].longitude) == false) && (std::isnan(GNSS_Data[i].altitude) == false)){
			operational_GNSS_units++;
			operational_GNSS = i;
		}
	}

	// Variance Testing
	if(operational_GNSS_units == 3){
    std::vector<std::vector<double>> VarVec; 
	VarVec = VarianceGNSS(20); // Set window length as argument
    for(auto it = VarVec.begin(); it != VarVec.end(); ++it){
		for(auto it2 = (*it).begin(); it2 != (*it).end(); ++ it2){
			int indexUnit = it2 - (*it).begin();
			int frozen = 0;
			int highvars = 0;
			
			if (std::isnan(VarVec[indexUnit][0]) || std::isnan(VarVec[indexUnit][1]) || std::isnan(VarVec[indexUnit][2])){
				goto nan_skipzone_gnss;}
			// Frozen Signal Testing
			if(doubleEqual(VarVec[indexUnit][0], 0.0, 1E-15)){
				GNNS_SignalHealth[indexUnit].FrozenSignal = true;
				GNSS_Data[indexUnit].deactivated = true;
				frozen++;
				ROS_WARN("Frozen latitude signal detected! GNSS unit %d",indexUnit+1);
			}
			if(doubleEqual(VarVec[indexUnit][1], 0.0, 1E-15)){
				GNNS_SignalHealth[indexUnit].FrozenSignal = true;
				GNSS_Data[indexUnit].deactivated = true;
				frozen++;
				ROS_WARN("Frozen longitude signal detected! GNSS unit %d",indexUnit +1);
			}
			if(doubleEqual(VarVec[indexUnit][2], 0.0, 1E-15)){
				GNNS_SignalHealth[indexUnit].FrozenSignal = true;
				GNSS_Data[indexUnit].deactivated = true;
				frozen++;
				ROS_WARN("Frozen altitude signal detected! GNSS unit %d",indexUnit +1);
			}
			// Reset boolean if signal has been classed as frozen, but has recently unfrozen. 
			if(GNNS_SignalHealth[indexUnit].FrozenSignal == true){
				if(frozen == 0){
					GNNS_SignalHealth[indexUnit].FrozenSignal = false;
				}
			}
			
			if(VarVec[indexUnit][0] > GNSS_max_var_horizontal){
				GNNS_SignalHealth[indexUnit].HighVarSignal = true;
				GNSS_Data[indexUnit].deactivated = true;
				highvars++;
				ROS_WARN("High variance latitude signal detected! GNSS unit %d",indexUnit+1);
			}
			if(VarVec[indexUnit][1] > GNSS_max_var_horizontal){
				GNNS_SignalHealth[indexUnit].HighVarSignal = true;
				GNSS_Data[indexUnit].deactivated = true;
				highvars++;
				ROS_WARN("High variance longitude signal detected! GNSS unit %d",indexUnit +1);
			}
			if(VarVec[indexUnit][2] > GNSS_max_var_vertical){ // Altitude is more noisy due to higher VDOP 
				GNNS_SignalHealth[indexUnit].HighVarSignal = true;
				GNSS_Data[indexUnit].deactivated = true;
				highvars++;
				ROS_WARN("High variance altitude signal detected! GNSS unit %d",indexUnit +1);}
			// Reset boolean if signal has been classed as under high variance, but has recently recovered. 
			if(GNNS_SignalHealth[indexUnit].HighVarSignal == true){
				if(highvars == 0){
					GNNS_SignalHealth[indexUnit].HighVarSignal = false;
				}
			}
		}
    }}
    nan_skipzone_gnss:
    
    // Check CUSUM results if any S-values are beyond thresholds:
    for(auto iter = GNSS_Residual_Data.begin(); iter != GNSS_Residual_Data.end(); ++iter){
		if(std::isnan((*iter).Vec1.measurement) == false){
			//ROS_WARN("S_high, S_low GPS1: %f %f", (*iter).Vec1.S_high, (*iter).Vec1.S_low);
			if((*iter).Vec1.S_high > GNSS_threshold_high){
				ROS_WARN("CUSUM test has exceeded the positive threshold! Vec1, unit %d",(*iter).unit);
				}
			if((*iter).Vec1.S_low < GNSS_threshold_low){
				ROS_WARN("CUSUM test has sunk below negative threshold! Vec1, unit %d",(*iter).unit);
				}
			if((*iter).Vec2.S_high > GNSS_threshold_high){
				ROS_WARN("CUSUM test has exceeded the positive threshold! Vec2 unit %d",(*iter).unit);
				}
			if((*iter).Vec2.S_low < GNSS_threshold_low){
				ROS_WARN("CUSUM test has sunk below negative threshold! Vec2, unit %d",(*iter).unit);
				}
			if((*iter).Vec3.S_high > 10*GNSS_threshold_high){
				ROS_WARN("CUSUM test has exceeded the positive threshold! Vec3, unit %d",(*iter).unit);
				}
			if((*iter).Vec3.S_low < 10*GNSS_threshold_low){
				ROS_WARN("CUSUM test has sunk below negative threshold! Vec3, unit %d",(*iter).unit);
				}
			}
	}


	// Check if there are enough operational units to conduct voting
	operational_GNSS_units = 0;
	for(int i = 0; i < GNSS_Data.size(); i++){
		if(GNSS_Data[i].deactivated == false  && (std::isnan(GNSS_Data[i].latitude) == false) && (std::isnan(GNSS_Data[i].longitude) == false) && (std::isnan(GNSS_Data[i].altitude) == false)){
			operational_GNSS_units++;
			operational_GNSS = i;
		}
	}
	
	// Either perform voting algorithm on measurements, if the number of working units exceeds two (2),
	// or print warning to inform the user no voting can be conducted. 
	if(operational_GNSS_units > 2){
		int lat_vote = vote(GNSS_Data[0].latitude,GNSS_Data[1].latitude,GNSS_Data[2].latitude, voting_threshold_horizontal);
		int lon_vote = vote(GNSS_Data[0].longitude,GNSS_Data[1].longitude,GNSS_Data[2].longitude,voting_threshold_horizontal);
		int alt_vote = vote(GNSS_Data[0].altitude,GNSS_Data[1].altitude,GNSS_Data[2].altitude,voting_threshold_vertical);
		
		
		std::vector<int> votes{lat_vote,lon_vote,alt_vote};
		GNSS_voting_warning_printed = false;

		for(int j = 0; j <= 2; j++){
			switch(votes[j]){
				case 0: {GNNS_SignalHealth[0].DeviationVoting = false;
						 GNNS_SignalHealth[1].DeviationVoting = false;
						 GNNS_SignalHealth[2].DeviationVoting = false;
						 break;}
				case 1: {GNNS_SignalHealth[0].DeviationVoting = true;
						 GNSS_Data[0].deactivated = true;
						 ROS_WARN("GNSS unit 1 voted out!");
						 break;}
				case 2: {GNNS_SignalHealth[1].DeviationVoting = true;
						 GNSS_Data[1].deactivated = true;
						 ROS_WARN("GNSS unit 2 voted out!");
						 break;}
				case 3: {GNNS_SignalHealth[2].DeviationVoting = true;
						 GNSS_Data[2].deactivated = true;
						 ROS_WARN("GNSS unit 3 voted out!");
						 break;}
			}
		}
		
		for(int i = 0; i <= 2; i++){
		// If a deactivated unit is now in harmony with the rest, it will be reactivated
			if(GNSS_votes_last[i] != 0 && votes[i] != GNSS_votes_last[i]){
				if(std::find(votes.begin(), votes.end(), GNSS_votes_last[i]) != votes.end()){
					GNNS_SignalHealth[GNSS_votes_last[i]].DeviationVoting = false;
					GNSS_Data[GNSS_votes_last[i]].deactivated = false; }
				}
			// Signal a flag if there has been a recent deactivation
			if(((std::find(votes.begin(), votes.end(),(i+1)) != votes.end()) == true) && ((std::find(GNSS_votes_last.begin(), GNSS_votes_last.end(),(i+1)) != GNSS_votes_last.end())== false)){
				GNSS_voted_out = true;
				GNSS_lp_timelapsed = 0.0;
				GNSS_time_last_lp = ros::Time::now().toSec();
				}
			}
		
		GNSS_votes_last = votes;
	}else{
		if(GNSS_voting_warning_printed == false){
			ROS_WARN("Not enough operational GNSS units to perform voting!");
			GNSS_voting_warning_printed = true;
		}
	}
	if(deactivate == 0){
		GNSS_Data[0].deactivated == false;
	}else if(deactivate == 1){
		GNSS_Data[0].deactivated == true;
		if(GNSS_voted_out == false){
			GNSS_voted_out == true;
			GNSS_lp_timelapsed = 0;
		}
	}
	
	// If two or more units are operational, the measurement should be averaged
	if(operational_GNSS_units >= 2){
		gnss_treated = averageGNSS();}
	else{std::vector<double> gnss_treated{GNSS_Data[operational_GNSS].latitude,GNSS_Data[operational_GNSS].longitude,GNSS_Data[operational_GNSS].altitude};}
	
	
	// If a GNSS unit has been voted out, ensure that the measurements are low-passed to prevent step changes
	if(GNSS_voted_out == true && GNSS_lp_timelapsed <= GNSS_lowpass_time){
		GNSS_lp_timelapsed += ros::Time::now().toSec() - GNSS_time_last_lp;
		for(int i = 0; i <= 2; i++){ 
			GNSS_LP_output_last[i] = LowPassFilter(gnss_treated[i],GNSS_LP_output_last[i]);}
		gnss_treated = GNSS_LP_output_last;
		ROS_WARN("LP started");
	} // Reset LP filter if sufficient time has passed
	else if(GNSS_voted_out == true && GNSS_lp_timelapsed > GNSS_lowpass_time){
		GNSS_voted_out = false;
		GNSS_LP_output_last = {0.0,0.0,0.0};
		ROS_WARN("LP Filter reset");
	}

	// Report sensor ready for duty in case failure mode has vanquished
	for(int i = 0; i < NUM_GNSS_UNITS; i++){
		if(GNSS_Data[i].deactivated == true){
			if(sensor_reactivator(GNNS_SignalHealth, i)){
				GNSS_Data[i].deactivated = false;}
		}
	}
	
	// Publish Processed Measurements
	gnssPub(gnss_treated);
}

void SigBlock::imuFTC(){
	// Send IMU message to residual generating EKF, which CUSUM will use later
	for(int i = 0; i < IMU_Data.size(); i++){
		if((std::isnan(IMU_Data[i].rollRate) == false) && (IMU_Data[i].published == false)){
			custom_msgs::inertialSensor residual_calculation;
			residual_calculation.roll = IMU_Data[i].roll;
			residual_calculation.pitch = IMU_Data[i].pitch;
			residual_calculation.yaw = IMU_Data[i].yaw;
			residual_calculation.rollRate = IMU_Data[i].rollRate;
			residual_calculation.pitchRate = IMU_Data[i].pitchRate;
			residual_calculation.yawRate = IMU_Data[i].yawRate;
			residual_calculation.imuUnit = i + 1;
			switch(i){
				case 0: {imu_residual_pub1.publish(residual_calculation); break;}
				case 1: {imu_residual_pub2.publish(residual_calculation); break;}
				case 2: {imu_residual_pub3.publish(residual_calculation); break;}
			}
			IMU_Data[i].published = true;
		}
   }
   
	//// Variance Testing
	int operational_IMU_units = 0;
	for(int i = 0; i < IMU_Data.size(); i++){
		if((IMU_Data[i].deactivated == false) && (std::isnan(IMU_Data[i].roll) == false) && (std::isnan(IMU_Data[i].rollRate) == false)){
			operational_IMU_units++;
			operational_IMU = i;
		}
	}
	
	if(operational_IMU_units == 3){
    std::vector<std::vector<double>> VarVec; 
	VarVec = VarianceIMU(20); // Set window length as argument
	if(VarVec[0].size() != 0 && VarVec[1].size() != 0 && VarVec[2].size() != 0){
    for(int indexUnit = 0; indexUnit <= 2; indexUnit++){
			int frozen = 0;
			int highvars = 0;
			if ( std::isnan(VarVec[indexUnit][0]) || std::isnan(VarVec[indexUnit][1]) || std::isnan(VarVec[indexUnit][2]) || std::isnan(VarVec[indexUnit][3]) || std::isnan(VarVec[indexUnit][4]) || std::isnan(VarVec[indexUnit][5])){
				goto nan_skipzone;}
			// Frozen Signal Testing
			if(doubleEqual(VarVec[indexUnit][0], 0.0, 1E-15)){
				IMU_SignalHealth[indexUnit].FrozenSignal = true;
				frozen++;
				ROS_WARN("Frozen roll signal detected! IMU unit %d",indexUnit+1);
			}
			if(doubleEqual(VarVec[indexUnit][1], 0.0, 1E-15)){
				IMU_SignalHealth[indexUnit].FrozenSignal = true;
				frozen++;
				ROS_WARN("Frozen pitch signal detected! IMU unit %d",indexUnit +1);
			}
			if(doubleEqual(VarVec[indexUnit][2], 0.0, 1E-15)){
				IMU_SignalHealth[indexUnit].FrozenSignal = true;
				frozen++;
				ROS_WARN("Frozen yaw signal detected! IMU unit %d",indexUnit +1);
			}
			if(doubleEqual(VarVec[indexUnit][3], 0.0, 1E-15)){
				IMU_SignalHealth[indexUnit].FrozenSignal = true;
				frozen++;
				ROS_WARN("Frozen roll rate signal detected! IMU unit %d",indexUnit +1);
			}
			if(doubleEqual(VarVec[indexUnit][4], 0.0, 1E-15)){
				IMU_SignalHealth[indexUnit].FrozenSignal = true;
				frozen++;
				ROS_WARN("Frozen pitch rate signal detected! IMU unit %d",indexUnit +1);
			}
			if(doubleEqual(VarVec[indexUnit][5], 0.0, 1E-15)){
				IMU_SignalHealth[indexUnit].FrozenSignal = true;
				frozen++;
				ROS_WARN("Frozen yaw rate signal detected! IMU unit %d",indexUnit +1);
			}
			// Reset boolean if signal has been classed as frozen, but has recently unfrozen. 
			if(IMU_SignalHealth[indexUnit].FrozenSignal == true){
				if(frozen == 0){
					IMU_SignalHealth[indexUnit].FrozenSignal = false;
				}
			}
			
			if(VarVec[indexUnit][0] > IMU_max_var_angles){
				IMU_SignalHealth[indexUnit].HighVarSignal = true;
				highvars++;
				ROS_WARN("High variance roll signal detected! IMU unit %d",indexUnit+1);
			}
			if(VarVec[indexUnit][1] > IMU_max_var_angles){
				IMU_SignalHealth[indexUnit].HighVarSignal = true;
				highvars++;
				ROS_WARN("High variance pitch signal detected! IMU unit %d",indexUnit +1);
			}
			if(VarVec[indexUnit][2] > IMU_max_var_angles){
				IMU_SignalHealth[indexUnit].HighVarSignal = true;
				highvars++;
				highvars++;
				ROS_WARN("High variance yaw signal detected! IMU unit %d",indexUnit +1);}
			if(VarVec[indexUnit][3] > IMU_max_var_rates){
				IMU_SignalHealth[indexUnit].HighVarSignal = true;
				highvars++;
				ROS_WARN("High variance roll rate signal detected! IMU unit %d",indexUnit +1);}
			if(VarVec[indexUnit][4] > IMU_max_var_rates){
				IMU_SignalHealth[indexUnit].HighVarSignal = true;
				highvars++;
				ROS_WARN("High variance pitch rate signal detected! IMU unit %d",indexUnit +1);}
			if(VarVec[indexUnit][5] > IMU_max_var_rates){
				IMU_SignalHealth[indexUnit].HighVarSignal = true;
				highvars++;
				ROS_WARN("High variance yaw rate signal detected! IMU unit %d",indexUnit +1);}
			
			// Reset boolean if signal has been classed as under high variance, but has recently recovered. 
			if(IMU_SignalHealth[indexUnit].HighVarSignal == true){
				if(highvars == 0){
					IMU_SignalHealth[indexUnit].HighVarSignal = false;
				}
			}
		}
    }}
    nan_skipzone:
	for(int i = 0; i < IMU_Data.size(); i++){
		if((std::isnan(IMU_Data[i].roll) == false) && (std::isnan(IMU_Data[i].pitch) == false) && (std::isnan(IMU_Data[i].yaw) == false)){
			custom_msgs::inertialSensor outlier_check;
			outlier_check.roll = IMU_Data[i].roll;
			outlier_check.pitch = IMU_Data[i].pitch;
			outlier_check.yaw = IMU_Data[i].yaw;
			outlier_check.rollRate = IMU_Data[i].rollRate;
			outlier_check.pitchRate = IMU_Data[i].pitchRate;
			outlier_check.yawRate = IMU_Data[i].yawRate;
			outlier_check.imuUnit = i + 1;
			imu_outlier_pub.publish(outlier_check);
		}
    }
    
    
    // Check CUSUM results if any S-values are beyond thresholds:
    for(auto iter = IMU_Residual_Data.begin(); iter != IMU_Residual_Data.end(); ++iter){
		if(std::isnan((*iter).Vec1.measurement) == false){
			//ROS_WARN("S_high, S_low GPS1: %f %f", (*iter).Vec1.S_high, (*iter).Vec1.S_low);
			if((*iter).Vec1.S_high > IMU_threshold_high){
				ROS_WARN("CUSUM test has exceeded the positive threshold! IMU Roll, unit %d",(*iter).unit);
				}
			if((*iter).Vec1.S_low < IMU_threshold_low){
				ROS_WARN("CUSUM test has sunk below negative threshold! IMU Roll, unit %d",(*iter).unit);
				}
			if((*iter).Vec2.S_high > IMU_threshold_high){
				ROS_WARN("CUSUM test has exceeded the positive threshold! IMU Pitch, unit %d",(*iter).unit);
				}
			if((*iter).Vec2.S_low < IMU_threshold_low){
				ROS_WARN("CUSUM test has sunk below negative threshold! IMU Pitch, unit %d",(*iter).unit);
				}
			if((*iter).Vec3.S_high > IMU_threshold_high){
				ROS_WARN("CUSUM test has exceeded the positive threshold! IMU Yaw, unit %d",(*iter).unit);
				}
			if((*iter).Vec3.S_low < IMU_threshold_low){
				ROS_WARN("CUSUM test has sunk below negative threshold! IMU Yaw, unit %d",(*iter).unit);
				}
			if((*iter).Vec4.S_high > IMU_threshold_high){
				ROS_WARN("CUSUM test has exceeded the positive threshold! IMU Roll Rate, unit %d",(*iter).unit);
				}
			if((*iter).Vec4.S_low < IMU_threshold_low){
				ROS_WARN("CUSUM test has sunk below negative threshold! IMU Roll Rate, unit %d",(*iter).unit);
				}
			if((*iter).Vec5.S_high > IMU_threshold_high){
				ROS_WARN("CUSUM test has exceeded the positive threshold! IMU Pitch Rate, unit %d",(*iter).unit);
				}
			if((*iter).Vec5.S_low < IMU_threshold_low){
				ROS_WARN("CUSUM test has sunk below negative threshold! IMU Pitch Rate, unit %d",(*iter).unit);
				}
			if((*iter).Vec6.S_high > IMU_threshold_high){
				ROS_WARN("CUSUM test has exceeded the positive threshold! IMU Yaw Rate, unit %d",(*iter).unit);
				}
			if((*iter).Vec6.S_low < IMU_threshold_low){
				ROS_WARN("CUSUM test has sunk below negative threshold! IMU Yaw Rate, unit %d",(*iter).unit);
				}
			}
	}


	// Check if there are enough operational units to conduct voting
	operational_IMU_units = 0;
	for(int i = 0; i < IMU_Data.size(); i++){
		if((IMU_Data[i].deactivated == false) && (std::isnan(IMU_Data[i].roll) == false) && (std::isnan(IMU_Data[i].rollRate) == false)){
			operational_IMU_units++;
			operational_IMU = i;
		}
	}
	
	// Either perform voting algorithm on measurements, if the number of working units exceeds two (2),
	// or print warning to inform the user no voting can be conducted. 
	if(operational_IMU_units > 2){
		int roll_vote = vote(IMU_Data[0].roll,IMU_Data[1].roll,IMU_Data[2].roll, voting_threshold_angles);
		int pitch_vote = vote(IMU_Data[0].pitch,IMU_Data[1].pitch,IMU_Data[2].pitch,voting_threshold_angles);
		int yaw_vote = vote(IMU_Data[0].yaw,IMU_Data[1].yaw,IMU_Data[2].yaw,voting_threshold_angles);
		int rollRate_vote = vote(IMU_Data[0].rollRate,IMU_Data[1].rollRate,IMU_Data[2].rollRate,voting_threshold_rates);
		int pitchRate_vote = vote(IMU_Data[0].pitchRate,IMU_Data[1].pitchRate,IMU_Data[2].pitchRate,voting_threshold_rates);
		int yawRate_vote = vote(IMU_Data[0].yawRate,IMU_Data[1].yawRate,IMU_Data[2].yawRate,voting_threshold_rates);
		
		std::vector<int> votes{roll_vote,pitch_vote,yaw_vote,rollRate_vote,pitchRate_vote,yawRate_vote};
		IMU_voting_warning_printed = false;
		
		// Analyze voting results
		for(int j = 0; j <= 2; j++){
			switch(votes[0]){
				case 0: {IMU_SignalHealth[0].DeviationVoting = false;
						 IMU_SignalHealth[1].DeviationVoting = false;
						 IMU_SignalHealth[2].DeviationVoting = false;
						 break;}
				case 1: {IMU_SignalHealth[0].DeviationVoting = true;
						 IMU_Data[0].deactivated = true;
						 ROS_WARN("IMU unit 1 voted out!");
						 break;}
				case 2: {IMU_SignalHealth[1].DeviationVoting = true;
						 IMU_Data[1].deactivated = true;
						 ROS_WARN("IMU unit 2 voted out!");
						 break;}
				case 3: {IMU_SignalHealth[2].DeviationVoting = true;
						 IMU_Data[2].deactivated = true;
						 ROS_WARN("IMU unit 3 voted out!");
						 break;}
			}
		}

		for(int i = 0; i <= 2; i++){
		// If a deactivated unit is now in harmony with the rest, it will be reactivated
			if(IMU_votes_last[i] != 0 && votes[i] != IMU_votes_last[i]){
				if(std::find(votes.begin(), votes.end(), IMU_votes_last[i]) != votes.end()){
					IMU_SignalHealth[IMU_votes_last[i]].DeviationVoting = false;
					IMU_Data[IMU_votes_last[i]].deactivated = false; }
				}
			// Signal a flag if there has been a recent deactivation
			if(((std::find(votes.begin(), votes.end(),(i+1)) != votes.end()) == true) && ((std::find(IMU_votes_last.begin(), IMU_votes_last.end(),(i+1)) != IMU_votes_last.end())== false)){
				IMU_voted_out = true;
				IMU_lp_timelapsed = 0.0;
				IMU_time_last_lp = ros::Time::now().toSec();
				}
			}
		
		IMU_votes_last = votes;
	}else{
		if(IMU_voting_warning_printed == false){
			ROS_WARN("Not enough operational IMU units to perform voting!");
			IMU_voting_warning_printed = true;
		}
	}
	
	// If two or more units are operational, the measurement should be averaged
	if(operational_IMU_units >= 2){
		imu_treated = averageIMU();}
	else{std::vector<double> publish{IMU_Data[operational_IMU].roll,IMU_Data[operational_IMU].pitch,IMU_Data[operational_IMU].yaw,IMU_Data[operational_IMU].rollRate,IMU_Data[operational_IMU].pitchRate,IMU_Data[operational_IMU].yawRate};}
	
	
	// If an IMU unit has been voted out, ensure that the measurements are low-passed to prevent step changes
	if(IMU_voted_out == true && IMU_lp_timelapsed <= IMU_lowpass_time){
		IMU_lp_timelapsed += ros::Time::now().toSec() - IMU_time_last_lp;
		for(int i = 0; i <= 2; i++){ 
			IMU_LP_output_last[i] = LowPassFilter(imu_treated[i],IMU_LP_output_last[i]);}
		imu_treated = IMU_LP_output_last;
		
	} // Reset LP filter if sufficient time has passed
	else if(IMU_voted_out == true && IMU_lp_timelapsed > IMU_lowpass_time){
		IMU_voted_out = false;
		IMU_LP_output_last = {0.0,0.0,0.0};
	}
	
	// Report sensor ready for duty in case failure mode has vanquished
	for(int i = 0; i < NUM_IMU_UNITS; i++){
		if(IMU_Data[i].deactivated == true){
			if(sensor_reactivator(IMU_SignalHealth, i)){
				IMU_Data[i].deactivated = false;}
			}
	}
	
	// Publish Consensus
	imuPub(imu_treated);
}

std::vector<double> SigBlock::averageGNSS(){
	std::vector<double> latitudes{GNSS_Data[0].latitude,GNSS_Data[1].latitude,GNSS_Data[2].latitude};
	std::vector<double> longitudes{GNSS_Data[0].longitude,GNSS_Data[1].longitude,GNSS_Data[2].longitude};
	std::vector<double> altitudes{GNSS_Data[0].altitude,GNSS_Data[1].altitude,GNSS_Data[2].altitude};
	std::vector<double> value_vector{0.0,0.0,0.0};
	int operational_units = NUM_GNSS_UNITS;
	
    if(GNSS_Data[0].deactivated == true){
		latitudes.erase(latitudes.begin());
		longitudes.erase(longitudes.begin());
		altitudes.erase(altitudes.begin());
		operational_units--;
	}
	else if(GNSS_Data[1].deactivated == true){
		latitudes.erase(latitudes.begin()+1);
		longitudes.erase(longitudes.begin()+1);
		altitudes.erase(altitudes.begin()+1);
		operational_units--;
	}
	else if(GNSS_Data[2].deactivated == true){
		latitudes.erase(latitudes.begin()+2);
		longitudes.erase(longitudes.begin()+2);
		altitudes.erase(altitudes.begin()+2);
		operational_units--;
	}
	if(operational_units >= 2){
		double latitude_avg = averageVector(latitudes);
		double longitude_avg = averageVector(longitudes);
		double altitude_avg = averageVector(altitudes);
	    std::vector<double> averages{latitude_avg,longitude_avg,altitude_avg};
	    value_vector = averages;
	} else{std::vector<double> only_functional{latitudes[0],longitudes[0],altitudes[0]};
		   value_vector = only_functional;}
    
    return value_vector;     
}

bool SigBlock::sensor_reactivator(std::vector<SIG_INFO> unit_health, int unit){
	if((unit_health[unit].LostSignal == false) && (unit_health[unit].FrozenSignal == false) && (unit_health[unit].HighVarSignal == false) && 
	(unit_health[unit].DriftingSignal == false) && (unit_health[unit].OutlierAlarm == false) && (unit_health[unit].ScalingError == false) && 
	(unit_health[unit].InvalidRange == false) && (unit_health[unit].DeviationVoting == false)){
		return true;}
	else{
		return false;
	}
	}

std::vector<double> SigBlock::averageIMU(){
	std::vector<double> roll{IMU_Data[0].roll,IMU_Data[1].roll,IMU_Data[2].roll};
	std::vector<double> pitch{IMU_Data[0].pitch,IMU_Data[1].pitch,IMU_Data[2].pitch};
	std::vector<double> yaw{IMU_Data[0].yaw,IMU_Data[1].yaw,IMU_Data[2].yaw};
	std::vector<double> rollRate{IMU_Data[0].rollRate,IMU_Data[1].rollRate,IMU_Data[2].rollRate};
	std::vector<double> pitchRate{IMU_Data[0].pitchRate,IMU_Data[1].pitchRate,IMU_Data[2].pitchRate};
	std::vector<double> yawRate{IMU_Data[0].yawRate,IMU_Data[1].yawRate,IMU_Data[2].yawRate};
	std::vector<double> value_vector{0.0,0.0,0.0,0.0,0.0,0.0};
	int operational_units = NUM_IMU_UNITS;

	
    if(IMU_Data[0].deactivated == true){
		roll.erase(roll.begin());
		pitch.erase(pitch.begin());
		yaw.erase(yaw.begin());
		rollRate.erase(rollRate.begin());
		pitchRate.erase(pitchRate.begin());
		yawRate.erase(yawRate.begin());
		operational_units--;
	}
	else if(IMU_Data[1].deactivated == true){
		roll.erase(roll.begin() + 1);
		pitch.erase(pitch.begin() + 1);
		yaw.erase(yaw.begin() + 1);
		rollRate.erase(rollRate.begin() + 1);
		pitchRate.erase(pitchRate.begin() + 1);
		yawRate.erase(yawRate.begin() + 1 );
		operational_units--;
	}
	else if(IMU_Data[2].deactivated == true){
		roll.erase(roll.begin() + 2);
		pitch.erase(pitch.begin() + 2);
		yaw.erase(yaw.begin() + 2);
		rollRate.erase(rollRate.begin() + 2);
		pitchRate.erase(pitchRate.begin() + 2);
		yawRate.erase(yawRate.begin() + 2);
		operational_units--;
	}
    
	if(operational_units >= 2){
		double roll_avg = averageVector(roll);
		double pitch_avg = averageVector(pitch);
		double yaw_avg = averageVector(yaw);    
		double rollRate_avg= averageVector(rollRate);
		double pitchRate_avg = averageVector(pitchRate);
		double yawRate_avg = averageVector(yawRate);
	    std::vector<double> averages{roll_avg,pitch_avg,yaw_avg,rollRate_avg,pitchRate_avg,yawRate_avg};
	    value_vector = averages;
	} else{std::vector<double> only_functional{roll[0],pitch[0],yaw[0],rollRate[0],pitchRate[0],yawRate[0]};
		   value_vector = only_functional;}
    return value_vector;     
}

// Variance Calculator GNSS

std::vector<std::vector<double>> SigBlock::VarianceGNSS(int samples){
    // This function finds the variance for each feature of all GNSS //
    // units, based on a window of [samples] amount of measurements  // 
    // IN: samples - amount of samples                               //
    // OUT: var_vec_all[X][Y] - 2D vector with variances, where X is //
    // unit number and Y is the feature of interest (lat,lon,alt)    //
    
    // Initialize vectors
    static std::vector<GNSS_DATA_VECTOR> LastSamples;
    std::vector<std::vector<double>> var_vec_all;
    std::vector<double> var_vec_individual;
    var_vec_individual.assign(3,nan(""));
    var_vec_all.assign(NUM_GNSS_UNITS,var_vec_individual);
    
    // Save sample in vector containing [samples] last GNSS measurements, if there's no NaNs present
    if(std::isnan(GNSS_Data[0].latitude) == false && std::isnan(GNSS_Data[1].latitude) == false && std::isnan(GNSS_Data[2].latitude) == false){
    LastSamples.insert(LastSamples.begin(), GNSS_Data);
    if(LastSamples.size() == (samples + 1)){
        LastSamples.pop_back();}
    }
    // Runs in case there are [samples] amount of data stored in the samples storage vector
    if(LastSamples.size() == samples){
        // Initialization of vectors local to the scope
        
        std::vector<double> mean_vec;
        std::vector<std::vector<double>> mean_vec_allUnits; 
        mean_vec.assign(3, 0.0);
        mean_vec_allUnits.assign(NUM_GNSS_UNITS, mean_vec);
        var_vec_individual.assign(3,0.0);
        var_vec_all.assign(NUM_GNSS_UNITS,var_vec_individual);
        
        // Finds the mean of the measurements of each unit
        
        for(auto iter = LastSamples.begin(); iter != LastSamples.end(); ++iter){
            for(auto iter2 = (*iter).begin(); iter2 != (*iter).end(); ++ iter2){
                
                int indexUnit = iter2 - (*iter).begin();
                int indexSample = iter - LastSamples.begin();

                // Sum up the last [samples] measurements
                mean_vec_allUnits[indexUnit][0] += (*iter2).latitude;
                mean_vec_allUnits[indexUnit][1] += (*iter2).longitude;
                mean_vec_allUnits[indexUnit][2] += (*iter2).altitude;

                // If summation is ended, Calculate mean of last [samples] measurements
                if((indexSample + 1) == samples){
                    mean_vec_allUnits[indexUnit][0] = mean_vec_allUnits[indexUnit][0]/samples; 
                    mean_vec_allUnits[indexUnit][1] = mean_vec_allUnits[indexUnit][1]/samples; 
                    mean_vec_allUnits[indexUnit][2] = mean_vec_allUnits[indexUnit][2]/samples;

                }
            }
        }

        
        // Calculate variance, and fill vector containing the variance of each measurement
        
        for(auto iter3 = LastSamples.begin(); iter3 != LastSamples.end(); ++iter3){
            for(auto iter4 = (*iter3).begin(); iter4 != (*iter3).end(); ++iter4){
                int indexUnit = iter4 - (*iter3).begin();
                    
                var_vec_all[indexUnit][0] += std::pow(((*iter4).latitude-mean_vec_allUnits[indexUnit][0]),2)/(samples-1);
                var_vec_all[indexUnit][1] += std::pow(((*iter4).longitude-mean_vec_allUnits[indexUnit][1]),2)/(samples-1);
                var_vec_all[indexUnit][2] += std::pow(((*iter4).altitude-mean_vec_allUnits[indexUnit][2]),2)/(samples-1);
            }
        }
    }          
    
    return var_vec_all;
}

// Variance Calculator IMU

std::vector<std::vector<double>> SigBlock::VarianceIMU(int samples){
    // This function finds the variance for each feature of relevant IMU //
    // units, based on a window of [samples] amount of measurements      // 
    // IN: samples - amount of samples                                   //
    // OUT: var_vec_all[X][Y] - 2D vector with variances, where X is     //
    // unit number and Y is the feature of interest (roll,rollrate..)    //
    
    // Initialize vectors
    static std::vector<IMU_DATA_VECTOR> LastSamples_IMU;
    std::vector<std::vector<double>> var_vec_all;
    std::vector<double> var_vec_individual;
    var_vec_individual.assign(6,nan(""));
    var_vec_all.assign(NUM_IMU_UNITS,var_vec_individual);
    
    // Save sample in vector containing [samples] last IMU measurements, if there's no NaNs present
    if(std::isnan(IMU_Data[0].roll) == false && std::isnan(IMU_Data[1].roll) == false && std::isnan(IMU_Data[2].roll) == false){
    LastSamples_IMU.insert(LastSamples_IMU.begin(), IMU_Data);
    if(LastSamples_IMU.size() == (samples + 1)){
        LastSamples_IMU.pop_back();
    	}
	}
    // Runs in case there are [samples] amount of data stored in the samples storage vector
    if(LastSamples_IMU.size() == samples){
        // Initialization of vectors local to the scope
        
        std::vector<double> mean_vec;
        std::vector<std::vector<double>> mean_vec_allUnits; 
        mean_vec.assign(6, 0.0);
        mean_vec_allUnits.assign(NUM_IMU_UNITS, mean_vec);
        var_vec_individual.assign(6,0.0);
        var_vec_all.assign(NUM_IMU_UNITS,var_vec_individual);
        
        // Finds the mean of the measurements of each unit
        
        for(auto iter = LastSamples_IMU.begin(); iter != LastSamples_IMU.end(); ++iter){
            for(auto iter2 = (*iter).begin(); iter2 != (*iter).end(); ++ iter2){
                
                int indexUnit = iter2 - (*iter).begin();
                int indexSample = iter - LastSamples_IMU.begin();

                // Sum up the last [samples] measurements
                mean_vec_allUnits[indexUnit][0] += (*iter2).roll;
                mean_vec_allUnits[indexUnit][1] += (*iter2).pitch;
                mean_vec_allUnits[indexUnit][2] += (*iter2).yaw;
                mean_vec_allUnits[indexUnit][3] += (*iter2).rollRate;
                mean_vec_allUnits[indexUnit][4] += (*iter2).pitchRate;
                mean_vec_allUnits[indexUnit][5] += (*iter2).yawRate;

                // If summation is ended, Calculate mean of last [samples] measurements
                if((indexSample + 1) == samples){
                    mean_vec_allUnits[indexUnit][0] = mean_vec_allUnits[indexUnit][0]/samples; 
                    mean_vec_allUnits[indexUnit][1] = mean_vec_allUnits[indexUnit][1]/samples; 
                    mean_vec_allUnits[indexUnit][2] = mean_vec_allUnits[indexUnit][2]/samples;
                    mean_vec_allUnits[indexUnit][3] = mean_vec_allUnits[indexUnit][3]/samples;
                    mean_vec_allUnits[indexUnit][4] = mean_vec_allUnits[indexUnit][4]/samples;
                    mean_vec_allUnits[indexUnit][5] = mean_vec_allUnits[indexUnit][5]/samples;

                }
            }
        }

        
        // Calculate variance, and fill vector containing the variance of each measurement
        
        for(auto iter3 = LastSamples_IMU.begin(); iter3 != LastSamples_IMU.end(); ++iter3){
            for(auto iter4 = (*iter3).begin(); iter4 != (*iter3).end(); ++iter4){
                int indexUnit = iter4 - (*iter3).begin();
                    
                var_vec_all[indexUnit][0] += std::pow(((*iter4).roll-mean_vec_allUnits[indexUnit][0]),2)/(samples-1);
                var_vec_all[indexUnit][1] += std::pow(((*iter4).pitch-mean_vec_allUnits[indexUnit][1]),2)/(samples-1);
                var_vec_all[indexUnit][2] += std::pow(((*iter4).yaw-mean_vec_allUnits[indexUnit][2]),2)/(samples-1);
                var_vec_all[indexUnit][3] += std::pow(((*iter4).rollRate-mean_vec_allUnits[indexUnit][3]),2)/(samples-1);
                var_vec_all[indexUnit][4] += std::pow(((*iter4).pitchRate-mean_vec_allUnits[indexUnit][4]),2)/(samples-1);
                var_vec_all[indexUnit][5] += std::pow(((*iter4).yawRate-mean_vec_allUnits[indexUnit][5]),2)/(samples-1);}
        }
    }

    return var_vec_all;
}


// Timestamp Checker

template<typename T, typename A>
void SigBlock::timestampChecker(std::vector<T,A> const& data_vec, int sensor_type){    
// Takes the latest timestamp of the each unit, and compares the time difference between each unit and the clock of ROS
// If the delay exceeds some maximum MaxDiff, the corresponding unit will be flagged for loss of signal. 
    std::vector<double>  TimestampVector{0.0,0.0,0.0};
    
    if(sensor_type == 0){
    for(int i = 0; i < NUM_GNSS_UNITS; i++){
        TimestampVector[i] = data_vec[i].timestamp_double;
    }}
    else if(sensor_type == 1){
    for(int i = 0; i < NUM_IMU_UNITS; i++){
        TimestampVector[i] = data_vec[i].timestamp_double;
    }}
    
    // Finds the maximal element of the timestamp vector => unit # for the latest measurement
    double MaxDiff = 4.0;
    double TimeNow = ros::Time::now().toSec();

    int index = 0;
    for (auto iter = TimestampVector.begin(); iter != TimestampVector.end(); ++iter){
        if((TimeNow - (*iter)) > MaxDiff){
            switch(index){
                case 0: {
                        GNNS_SignalHealth[0].LostSignal = true;
                        GNSS_Data[0].deactivated = true;
                        ROS_WARN("Abnormally long signal delay from GNSS unit #%i! Presumed loss of signal",index+1);
                        break;}
                case 1: {
                        GNNS_SignalHealth[1].LostSignal = true;
                        GNSS_Data[1].deactivated = true;
                        ROS_WARN("Abnormally long signal delay from GNSS unit #%i! Presumed loss of signal",index+1);
                        break;}
                case 2: {
                        GNNS_SignalHealth[2].LostSignal = true;
                        GNSS_Data[2].deactivated = true;
                        ROS_WARN("Abnormally long signal delay from GNSS unit #%i! Presumed loss of signal",index+1);
                        break;}
            }
            index++;
        }
    }
}


// Voting Algorithm (universal for both GNSS and IMU)

int SigBlock::vote(double &reading1,double &reading2,double &reading3, double tolerance){
	// If a measurement of a sensor deviates more than a specific amount (specified by the tolerance), it is voted out. Requires 3 or more units to be operational for voting. 
    if ((fabs(reading1 - reading2)) < tolerance && (fabs(reading1 - reading3) < tolerance)){
        return 0;
    }
    else if ((fabs(reading1 - reading2) < tolerance) && !(fabs(reading1 - reading3) < tolerance)){
        return 3;
    }
    else if (!(fabs(reading1 - reading2) < tolerance) && (fabs(reading1 - reading3) < tolerance)){
        return 2;
    }
    else if ((fabs(reading2 - reading3) < tolerance) && !(fabs(reading1 - reading3) < tolerance)){
        return 1;
    }
}

double SigBlock::LowPassFilter(double Input, double Output_last){
	// Simple low pass filter used in case of sensor deactivation
	// out(k) = (1-beta_LP)*out(k-1) + (beta_LP*in(k))) = out(n-1) - (beta_LP*(out(k-1)-in(k)))
	// Where beta_LP is a tuning factor akin to the time constant in the Laplace LP filter, with a range from 0 to 1
	double Output_now = Output_last - (beta_LP * (Output_last - Input));
	return Output_now;
}


std::vector<double> SigBlock::CUSUM_test(double residual, double u_H1,double u_H2, double sigma_null, double sigma_H1, double sigma_H2, std::vector<double> S_values){

	boost::math::normal N0(0.0, sigma_null);
	boost::math::normal N1(u_H1, sigma_H1);
    boost::math::normal N2(u_H2, sigma_H2);

	double Prob0_null_hypothesis = pdf(N0,residual);
	double Prob_H1 = pdf(N1,residual);
	double Prob_H2 = pdf(N2,residual);

	double S_high = S_values[0];
	double S_low = S_values[1];

	//ROS_WARN("Prob0 %f Prob1 %f", Prob0, Prob1);
	// Positive Direction
	S_high = std::max(0.0,S_high + std::log10(Prob_H1/Prob0_null_hypothesis));
	
	// Negative Direction
	S_low = std::min(0.0,-(S_low + std::log10(Prob_H2/Prob0_null_hypothesis)));
	//S_low = std::min(0.0,S_low + std::log10(Prob_H2/Prob0_null_hypothesis));
	//ROS_WARN("S_high %f S_low %f", S_high, S_low);
	//ROS_WARN("Prob0 ProbH1 ProbH2 residual %f %f %f %f", Prob0_null_hypothesis, Prob_H1, Prob_H2, residual);
	S_values[0] = S_high;
	S_values[1] = S_low;
	return S_values;
}

